@@ -46,6 +46,7 @@ void appctx_shut(struct appctx *appctx);
 struct appctx *appctx_new_on(struct applet *applet, struct sedesc *sedesc, int thr);
 int appctx_finalize_startup(struct appctx *appctx, struct proxy *px, struct buffer *input);
 void appctx_free_on_early_error(struct appctx *appctx);
+void appctx_free(struct appctx *appctx);
 
 static inline struct appctx *appctx_new_here(struct applet *applet, struct sedesc *sedesc)
 {
@@ -88,22 +89,6 @@ static inline void __appctx_free(struct appctx *appctx)
 	_HA_ATOMIC_DEC(&nb_applets);
 }
 
-static inline void appctx_free(struct appctx *appctx)
-{
-	/* The task is supposed to be run on this thread, so we can just
-	 * check if it's running already (or about to run) or not
-	 */
-	if (!(appctx->t->state & (TASK_QUEUED | TASK_RUNNING)))
-		__appctx_free(appctx);
-	else {
-		/* if it's running, or about to run, defer the freeing
-		 * until the callback is called.
-		 */
-		appctx->state |= APPLET_WANT_DIE;
-		task_wakeup(appctx->t, TASK_WOKEN_OTHER);
-	}
-}
-
 /* wakes up an applet when conditions have changed. We're using a macro here in
  * order to retrieve the caller's place.
  */
@@ -141,29 +126,45 @@ static inline void applet_have_no_more_data(struct appctx *appctx)
 }
 
 /* The applet indicates that it's ready to consume data from the stream's
- * output buffer.
+ * output buffer. Rely on the corresponding SE function
  */
 static inline void applet_will_consume(struct appctx *appctx)
 {
-	se_fl_clr(appctx->sedesc, SE_FL_WONT_CONSUME);
+	se_will_consume(appctx->sedesc);
 }
 
 /* The applet indicates that it's not willing to consume data from the stream's
- * output buffer.
+ * output buffer.  Rely on the corresponding SE function
  */
 static inline void applet_wont_consume(struct appctx *appctx)
 {
-	se_fl_set(appctx->sedesc, SE_FL_WONT_CONSUME);
+	se_wont_consume(appctx->sedesc);
 }
 
 /* The applet indicates that it's willing to consume data from the stream's
  * output buffer, but that there's not enough, so it doesn't want to be woken
- * up until more are presented.
+ * up until more are presented. Rely on the corresponding SE function
  */
 static inline void applet_need_more_data(struct appctx *appctx)
 {
-	se_fl_clr(appctx->sedesc, SE_FL_WONT_CONSUME);
-	se_fl_set(appctx->sedesc, SE_FL_WAIT_DATA);
+	se_need_more_data(appctx->sedesc);
+}
+
+/* The applet indicates that it does not expect data from the opposite endpoint.
+ * This way the stream know it should not trigger read timeout on the other
+ * side.
+ */
+static inline void applet_expect_no_data(struct appctx *appctx)
+{
+	se_fl_set(appctx->sedesc, SE_FL_EXP_NO_DATA);
+}
+
+/* The applet indicates that it expects data from the opposite endpoint. This
+ * way the stream know it may trigger read timeout on the other side.
+ */
+static inline void applet_expect_data(struct appctx *appctx)
+{
+	se_fl_clr(appctx->sedesc, SE_FL_EXP_NO_DATA);
 }
 
 /* writes chunk <chunk> into the input channel of the stream attached to this
@@ -177,7 +178,7 @@ static inline int applet_putchk(struct appctx *appctx, struct buffer *chunk)
 
 	ret = ci_putchk(sc_ic(se->sc), chunk);
 	if (ret == -1)
-		sc_need_room(se->sc);
+		sc_need_room(se->sc, chunk->data);
 
 	return ret;
 }
@@ -193,7 +194,7 @@ static inline int applet_putblk(struct appctx *appctx, const char *blk, int len)
 
 	ret = ci_putblk(sc_ic(se->sc), blk, len);
 	if (ret == -1)
-		sc_need_room(se->sc);
+		sc_need_room(se->sc, len);
 
 	return ret;
 }
@@ -210,7 +211,7 @@ static inline int applet_putstr(struct appctx *appctx, const char *str)
 
 	ret = ci_putstr(sc_ic(se->sc), str);
 	if (ret == -1)
-		sc_need_room(se->sc);
+		sc_need_room(se->sc, strlen(str));
 
 	return ret;
 }
@@ -226,7 +227,7 @@ static inline int applet_putchr(struct appctx *appctx, char chr)
 
 	ret = ci_putchr(sc_ic(se->sc), chr);
 	if (ret == -1)
-		sc_need_room(se->sc);
+		sc_need_room(se->sc, 1);
 
 	return ret;
 }
