@@ -378,7 +378,7 @@ static inline void channel_add_input(struct channel *chn, unsigned int len)
 	}
 	/* notify that some data was read */
 	chn->total += len;
-	chn->flags |= CF_READ_PARTIAL;
+	chn->flags |= CF_READ_EVENT;
 }
 
 static inline unsigned long long channel_htx_forward(struct channel *chn, struct htx *htx, unsigned long long bytes)
@@ -514,35 +514,23 @@ static inline int channel_may_recv(const struct channel *chn)
 /* Returns true if the channel's input is already closed */
 static inline int channel_input_closed(struct channel *chn)
 {
-	return ((chn->flags & CF_SHUTR) != 0);
+	return ((chn_prod(chn)->flags & (SC_FL_ABRT_DONE|SC_FL_EOS)) != 0);
 }
 
 /* Returns true if the channel's output is already closed */
 static inline int channel_output_closed(struct channel *chn)
 {
-	return ((chn->flags & CF_SHUTW) != 0);
+	return ((chn_cons(chn)->flags & SC_FL_SHUT_DONE) != 0);
 }
 
-/* Check channel timeouts, and set the corresponding flags. The likely/unlikely
- * have been optimized for fastest normal path. The read/write timeouts are not
- * set if there was activity on the channel. That way, we don't have to update
- * the timeout on every I/O. Note that the analyser timeout is always checked.
- */
-static inline void channel_check_timeouts(struct channel *chn)
+/* Check channel timeouts, and set the corresponding flags. */
+static inline void channel_check_timeout(struct channel *chn)
 {
-	if (likely(!(chn->flags & (CF_SHUTR|CF_READ_TIMEOUT|CF_READ_ACTIVITY|CF_READ_NOEXP))) &&
-	    unlikely(tick_is_expired(chn->rex, now_ms)))
-		chn->flags |= CF_READ_TIMEOUT;
-
-	if (likely(!(chn->flags & (CF_SHUTW|CF_WRITE_TIMEOUT|CF_WRITE_ACTIVITY))) &&
-	    unlikely(tick_is_expired(chn->wex, now_ms)))
-		chn->flags |= CF_WRITE_TIMEOUT;
-
-	if (likely(!(chn->flags & CF_ANA_TIMEOUT)) &&
-	    unlikely(tick_is_expired(chn->analyse_exp, now_ms)))
-		chn->flags |= CF_ANA_TIMEOUT;
+	if (likely(!(chn->flags & CF_READ_EVENT)) && unlikely(tick_is_expired(chn->analyse_exp, now_ms)))
+		chn->flags |= CF_READ_EVENT;
 }
 
+
 /* Erase any content from channel <buf> and adjusts flags accordingly. Note
  * that any spliced data is not affected since we may not have any access to
  * it.
@@ -560,22 +548,13 @@ static inline void channel_htx_erase(struct channel *chn, struct htx *htx)
 	channel_erase(chn);
 }
 
-/* marks the channel as "shutdown" ASAP for reads */
-static inline void channel_shutr_now(struct channel *chn)
-{
-	chn->flags |= CF_SHUTR_NOW;
-}
-
-/* marks the channel as "shutdown" ASAP for writes */
-static inline void channel_shutw_now(struct channel *chn)
-{
-	chn->flags |= CF_SHUTW_NOW;
-}
 
 /* marks the channel as "shutdown" ASAP in both directions */
 static inline void channel_abort(struct channel *chn)
 {
-	chn->flags |= CF_SHUTR_NOW | CF_SHUTW_NOW;
+	chn_prod(chn)->flags |= SC_FL_ABRT_WANTED;
+	chn_cons(chn)->flags |= SC_FL_SHUT_WANTED;
+	chn->flags |= CF_AUTO_CLOSE;
 	chn->flags &= ~CF_AUTO_CONNECT;
 }
 
@@ -998,8 +977,8 @@ static inline int ci_putstr(struct channel *chn, const char *str)
 static inline int co_getchr(struct channel *chn)
 {
 	/* closed or empty + imminent close = -2; empty = -1 */
-	if (unlikely((chn->flags & CF_SHUTW) || channel_is_empty(chn))) {
-		if (chn->flags & (CF_SHUTW|CF_SHUTW_NOW))
+	if (unlikely((chn_cons(chn)->flags & SC_FL_SHUT_DONE) || channel_is_empty(chn))) {
+		if (chn_cons(chn)->flags & (SC_FL_SHUT_DONE|SC_FL_SHUT_WANTED))
 			return -2;
 		return -1;
 	}
