@@ -425,12 +425,12 @@ static void delete_entry(struct cache_entry *del_entry)
 
 static inline struct shared_context *shctx_ptr(struct cache *cache)
 {
-	return (struct shared_context *)((unsigned char *)cache - ((struct shared_context *)NULL)->data);
+	return (struct shared_context *)((unsigned char *)cache -  offsetof(struct shared_context, data));
 }
 
 static inline struct shared_block *block_ptr(struct cache_entry *entry)
 {
-	return (struct shared_block *)((unsigned char *)entry - ((struct shared_block *)NULL)->data);
+	return (struct shared_block *)((unsigned char *)entry - offsetof(struct shared_block, data));
 }
 
 
@@ -536,17 +536,7 @@ cache_store_strm_deinit(struct stream *s, struct filter *filter)
 	/* Everything should be released in the http_end filter, but we need to do it
 	 * there too, in case of errors */
 	if (st && st->first_block) {
-		struct cache_entry *object = (struct cache_entry *)st->first_block->data;
-
 		shctx_lock(shctx);
-		if (!object->complete) {
-			/* The stream was closed but the 'complete' flag was not
-			 * set which means that cache_store_http_end was not
-			 * called. The stream must have been closed before we
-			 * could store the full answer in the cache.
-			 */
-			delete_entry(object);
-		}
 		shctx_row_dec_hot(shctx, st->first_block);
 		shctx_unlock(shctx);
 	}
@@ -960,7 +950,7 @@ static int http_check_vary_header(struct htx *htx, unsigned int *vary_signature)
  * "vary" on the accept-encoding value.
  * Returns 0 if we found a known encoding in the response, -1 otherwise.
  */
-static int set_secondary_key_encoding(struct htx *htx, unsigned int vary_signature, char *secondary_key)
+static int set_secondary_key_encoding(struct htx *htx, char *secondary_key)
 {
 	unsigned int resp_encoding_bitmap = 0;
 	const struct vary_hashing_information *info = vary_information;
@@ -970,11 +960,6 @@ static int set_secondary_key_encoding(struct htx *htx, unsigned int vary_signatu
 	unsigned int encoding_value;
 	struct http_hdr_ctx ctx = { .blk = NULL };
 
-	/* We must not set the accept encoding part of the secondary signature
-	 * if the response does not vary on 'Accept Encoding'. */
-	if (!(vary_signature & VARY_ACCEPT_ENCODING))
-		return 0;
-
 	/* Look for the accept-encoding part of the secondary_key. */
 	while (count < hash_info_count && info->value != VARY_ACCEPT_ENCODING) {
 		offset += info->hash_length;
@@ -1233,7 +1218,7 @@ enum act_return http_action_store_cache(struct act_rule *rule, struct proxy *px,
 	 * We will not cache a response that has an unknown encoding (not
 	 * explicitly supported in parse_encoding_value function). */
 	if (cache->vary_processing_enabled && vary_signature)
-		if (set_secondary_key_encoding(htx, vary_signature, object->secondary_key))
+		if (set_secondary_key_encoding(htx, object->secondary_key))
 		    goto out;
 
 	shctx_lock(shctx);
@@ -1481,15 +1466,16 @@ static void http_cache_io_handler(struct appctx *appctx)
 	res_htx = htx_from_buf(&res->buf);
 	total = res_htx->data;
 
-	if (unlikely(sc->state == SC_ST_DIS || sc->state == SC_ST_CLO))
+	if (unlikely(se_fl_test(appctx->sedesc, (SE_FL_EOS|SE_FL_ERROR|SE_FL_SHR|SE_FL_SHW))))
 		goto out;
 
-	if (res->flags & (CF_SHUTW|CF_SHUTR|CF_SHUTW_NOW))
-		appctx->st0 = HTX_CACHE_END;
+	/* Check if the input buffer is available. */
+	if (!b_size(&res->buf)) {
+		sc_need_room(sc, 0);
+		goto out;
+	}
 
 	if (appctx->st0 == HTX_CACHE_INIT) {
-		if (!co_data(req))
-			goto wait_request;
 		ctx->next = block_ptr(cache_ptr);
 		ctx->offset = sizeof(*cache_ptr);
 		ctx->sent = 0;
@@ -1497,12 +1483,6 @@ static void http_cache_io_handler(struct appctx *appctx)
 		appctx->st0 = HTX_CACHE_HEADER;
 	}
 
-	/* Check if the input buffer is available. */
-	if (!b_size(&res->buf)) {
-		sc_need_room(sc);
-		goto out;
-	}
-
 	if (appctx->st0 == HTX_CACHE_HEADER) {
 		/* Headers must be dump at once. Otherwise it is an error */
 		len = first->len - sizeof(*cache_ptr) - ctx->sent;
@@ -1533,7 +1513,7 @@ static void http_cache_io_handler(struct appctx *appctx)
 		if (len) {
 			ret = htx_cache_dump_msg(appctx, res_htx, len, HTX_BLK_UNUSED);
 			if (ret < len) {
-				sc_need_room(sc);
+				sc_need_room(sc, channel_htx_recv_max(res, res_htx) + 1);
 				goto out;
 			}
 		}
@@ -1543,16 +1523,13 @@ static void http_cache_io_handler(struct appctx *appctx)
 	if (appctx->st0 == HTX_CACHE_EOM) {
 		 /* no more data are expected. */
 		res_htx->flags |= HTX_FL_EOM;
-		res->flags |= CF_EOI;
 		se_fl_set(appctx->sedesc, SE_FL_EOI);
 		appctx->st0 = HTX_CACHE_END;
 	}
 
   end:
-	if (!(res->flags & CF_SHUTR) && appctx->st0 == HTX_CACHE_END) {
-		res->flags |= CF_READ_NULL;
-		sc_shutr(sc);
-	}
+	if (appctx->st0 == HTX_CACHE_END)
+		se_fl_set(appctx->sedesc, SE_FL_EOS);
 
   out:
 	total = res_htx->data - total;
@@ -1568,12 +1545,6 @@ static void http_cache_io_handler(struct appctx *appctx)
 	}
 	return;
 
-  wait_request:
-	/* Wait for the request before starting to deliver the response */
-	b_reset(&res->buf);
-	applet_need_more_data(appctx);
-	return;
-
   error:
 	/* Sent and HTTP error 500 */
 	b_reset(&res->buf);
@@ -1583,6 +1554,7 @@ static void http_cache_io_handler(struct appctx *appctx)
 	res_htx = htx_from_buf(&res->buf);
 
 	total = 0;
+	se_fl_set(appctx->sedesc, SE_FL_ERROR);
 	appctx->st0 = HTX_CACHE_END;
 	goto end;
 }
