@@ -47,6 +47,8 @@ struct bind_kw_list bind_keywords = {
 /* list of the temporarily limited listeners because of lack of resource */
 static struct mt_list global_listener_queue = MT_LIST_HEAD_INIT(global_listener_queue);
 static struct task *global_listener_queue_task;
+/* number of times an accepted connection resulted in maxconn being reached */
+ullong maxconn_reached = 0;
 __decl_thread(static HA_RWLOCK_T global_listener_rwlock);
 
 /* listener status for stats */
@@ -68,10 +70,10 @@ struct connection *accept_queue_pop_sc(struct accept_queue_ring *ring)
 	unsigned int pos, next;
 	struct connection *ptr;
 	struct connection **e;
+	uint32_t idx = _HA_ATOMIC_LOAD(&ring->idx);  /* (head << 16) + tail */
 
-	pos = ring->head;
-
-	if (pos == ring->tail)
+	pos = idx >> 16;
+	if (pos == (uint16_t)idx)
 		return NULL;
 
 	next = pos + 1;
@@ -93,7 +95,10 @@ struct connection *accept_queue_pop_sc(struct accept_queue_ring *ring)
 	*e = NULL;
 
 	__ha_barrier_store();
-	ring->head = next;
+	do {
+		pos = (next << 16) | (idx & 0xffff);
+	} while (unlikely(!HA_ATOMIC_CAS(&ring->idx, &idx, pos) && __ha_cpu_relax()));
+
 	return ptr;
 }
 
@@ -105,15 +110,17 @@ struct connection *accept_queue_pop_sc(struct accept_queue_ring *ring)
 int accept_queue_push_mp(struct accept_queue_ring *ring, struct connection *conn)
 {
 	unsigned int pos, next;
+	uint32_t idx = _HA_ATOMIC_LOAD(&ring->idx);  /* (head << 16) + tail */
 
-	pos = ring->tail;
 	do {
+		pos = (uint16_t)idx;
 		next = pos + 1;
 		if (next >= ACCEPT_QUEUE_SIZE)
 			next = 0;
-		if (next == ring->head)
+		if (next == (idx >> 16))
 			return 0; // ring full
-	} while (unlikely(!_HA_ATOMIC_CAS(&ring->tail, &pos, next)));
+		next |= (idx & 0xffff0000U);
+	} while (unlikely(!_HA_ATOMIC_CAS(&ring->idx, &idx, next) && __ha_cpu_relax()));
 
 	ring->entry[pos] = conn;
 	__ha_barrier_store();
@@ -142,19 +149,20 @@ struct task *accept_queue_process(struct task *t, void *context, unsigned int st
 
 		li = __objt_listener(conn->target);
 		_HA_ATOMIC_INC(&li->thr_conn[ti->ltid]);
-		ret = li->accept(conn);
+		ret = li->bind_conf->accept(conn);
 		if (ret <= 0) {
 			/* connection was terminated by the application */
 			continue;
 		}
 
 		/* increase the per-process number of cumulated sessions, this
-		 * may only be done once l->accept() has accepted the connection.
+		 * may only be done once l->bind_conf->accept() has accepted the
+		 * connection.
 		 */
-		if (!(li->options & LI_O_UNLIMITED)) {
+		if (!(li->bind_conf->options & BC_O_UNLIMITED)) {
 			HA_ATOMIC_UPDATE_MAX(&global.sps_max,
 			                     update_freq_ctr(&global.sess_per_sec, 1));
-			if (li->bind_conf && li->bind_conf->options & BC_O_USE_SSL) {
+			if (li->bind_conf->options & BC_O_USE_SSL) {
 				HA_ATOMIC_UPDATE_MAX(&global.ssl_max,
 				                     update_freq_ctr(&global.ssl_per_sec, 1));
 			}
@@ -195,8 +203,7 @@ static void accept_queue_deinit()
 	int i;
 
 	for (i = 0; i < global.nbthread; i++) {
-		if (accept_queue_rings[i].tasklet)
-			tasklet_free(accept_queue_rings[i].tasklet);
+		tasklet_free(accept_queue_rings[i].tasklet);
 	}
 }
 
@@ -204,19 +211,21 @@ REGISTER_POST_DEINIT(accept_queue_deinit);
 
 #endif // USE_THREAD
 
-/* Memory allocation and initialization of the per_thr field.
+/* Memory allocation and initialization of the per_thr field (one entry per
+ * bound thread).
  * Returns 0 if the field has been successfully initialized, -1 on failure.
  */
 int li_init_per_thr(struct listener *li)
 {
+	int nbthr = MIN(global.nbthread, MAX_THREADS_PER_GROUP);
 	int i;
 
 	/* allocate per-thread elements for listener */
-	li->per_thr = calloc(global.nbthread, sizeof(*li->per_thr));
+	li->per_thr = calloc(nbthr, sizeof(*li->per_thr));
 	if (!li->per_thr)
 		return -1;
 
-	for (i = 0; i < global.nbthread; ++i) {
+	for (i = 0; i < nbthr; ++i) {
 		MT_LIST_INIT(&li->per_thr[i].quic_accept.list);
 		MT_LIST_INIT(&li->per_thr[i].quic_accept.conns);
 
@@ -229,7 +238,7 @@ int li_init_per_thr(struct listener *li)
 /* helper to get listener status for stats */
 enum li_status get_li_status(struct listener *l)
 {
-	if (!l->maxconn || l->nbconn < l->maxconn) {
+	if (!l->bind_conf->maxconn || l->nbconn < l->bind_conf->maxconn) {
 		if (l->state == LI_LIMITED)
 			return LI_STATUS_WAITING;
 		else
@@ -320,7 +329,7 @@ void enable_listener(struct listener *listener)
 			 */
 			do_unbind_listener(listener);
 		}
-		else if (!listener->maxconn || listener->nbconn < listener->maxconn) {
+		else if (!listener->bind_conf->maxconn || listener->nbconn < listener->bind_conf->maxconn) {
 			listener->rx.proto->enable(listener);
 			listener_set_state(listener, LI_READY);
 		}
@@ -345,7 +354,7 @@ void stop_listener(struct listener *l, int lpx, int lpr, int lli)
 {
 	struct proxy *px = l->bind_conf->frontend;
 
-	if (l->options & LI_O_NOSTOP) {
+	if (l->bind_conf->options & BC_O_NOSTOP) {
 		/* master-worker sockpairs are never closed but don't count as a
 		 * job.
 		 */
@@ -582,7 +591,7 @@ int resume_listener(struct listener *l, int lpx, int lli)
 			goto end; /* failure to resume */
 	}
 
-	if (l->maxconn && l->nbconn >= l->maxconn) {
+	if (l->bind_conf->maxconn && l->nbconn >= l->bind_conf->maxconn) {
 		l->rx.proto->disable(l);
 		listener_set_state(l, LI_FULL);
 		goto done;
@@ -803,13 +812,84 @@ int create_listeners(struct bind_conf *bc, const struct sockaddr_storage *ss,
 	return 1;
 }
 
+/* Optionally allocates a new shard info (if si == NULL) for receiver rx and
+ * assigns it to it, or attaches to an existing one. If the rx already had a
+ * shard_info, it is simply returned. It is illegal to call this function with
+ * an rx that's part of a group that is already attached. Attaching means the
+ * shard_info's thread count and group count are updated so the rx's group is
+ * added to the shard_info's group mask. The rx are added to the members in the
+ * attachment order, though it must not matter. It is meant for boot time setup
+ * and is not thread safe. NULL is returned on allocation failure.
+ */
+struct shard_info *shard_info_attach(struct receiver *rx, struct shard_info *si)
+{
+	if (rx->shard_info)
+		return rx->shard_info;
+
+	if (!si) {
+		si = calloc(1, sizeof(*si));
+		if (!si)
+			return NULL;
+
+		si->ref = rx;
+	}
+
+	rx->shard_info = si;
+	BUG_ON (si->tgroup_mask & 1UL << (rx->bind_tgroup - 1));
+	si->tgroup_mask |= 1UL << (rx->bind_tgroup - 1);
+	si->nbgroups     = my_popcountl(si->tgroup_mask);
+	si->nbthreads   += my_popcountl(rx->bind_thread);
+	si->members[si->nbgroups - 1] = rx;
+	return si;
+}
+
+/* Detaches the rx from an optional shard_info it may be attached to. If so,
+ * the thread counts, group masks and refcounts are updated. The members list
+ * remains contiguous by replacing the current entry with the last one. The
+ * reference continues to point to the first receiver. If the group count
+ * reaches zero, the shard_info is automatically released.
+ */
+void shard_info_detach(struct receiver *rx)
+{
+	struct shard_info *si = rx->shard_info;
+	uint gr;
+
+	if (!si)
+		return;
+
+	rx->shard_info = NULL;
+
+	/* find the member slot this rx was attached to */
+	for (gr = 0; gr < MAX_TGROUPS && si->members[gr] != rx; gr++)
+		;
+
+	BUG_ON(gr == MAX_TGROUPS);
+
+	si->nbthreads   -= my_popcountl(rx->bind_thread);
+	si->tgroup_mask &= ~(1UL << (rx->bind_tgroup - 1));
+	si->nbgroups     = my_popcountl(si->tgroup_mask);
+
+	/* replace the member by the last one. If we removed the reference, we
+	 * have to switch to another one. It's always the first entry so we can
+	 * simply enforce it upon every removal.
+	 */
+	si->members[gr] = si->members[si->nbgroups];
+	si->members[si->nbgroups] = NULL;
+	si->ref = si->members[0];
+
+	if (!si->nbgroups)
+		free(si);
+}
+
 /* clones listener <src> and returns the new one. All dynamically allocated
  * fields are reallocated (name for now). The new listener is inserted before
  * the original one in the bind_conf and frontend lists. This allows it to be
  * duplicated while iterating over the current list. The original listener must
  * only be in the INIT or ASSIGNED states, and the new listener will only be
  * placed into the INIT state. The counters are always set to NULL. Maxsock is
- * updated. Returns NULL on allocation error.
+ * updated. Returns NULL on allocation error. The shard_info is never taken so
+ * that the caller can decide what to do with it depending on how it intends to
+ * clone the listener.
  */
 struct listener *clone_listener(struct listener *src)
 {
@@ -827,6 +907,7 @@ struct listener *clone_listener(struct listener *src)
 	}
 
 	l->rx.owner = l;
+	l->rx.shard_info = NULL;
 	l->state = LI_INIT;
 	l->counters = NULL;
 	l->extra_counters = NULL;
@@ -862,6 +943,7 @@ void __delete_listener(struct listener *listener)
 	if (listener->state == LI_ASSIGNED) {
 		listener_set_state(listener, LI_INIT);
 		LIST_DELETE(&listener->rx.proto_list);
+		shard_info_detach(&listener->rx);
 		listener->rx.proto->nb_receivers--;
 		_HA_ATOMIC_DEC(&jobs);
 		_HA_ATOMIC_DEC(&listeners);
@@ -887,14 +969,14 @@ void delete_listener(struct listener *listener)
  */
 int listener_backlog(const struct listener *l)
 {
-	if (l->backlog)
-		return l->backlog;
+	if (l->bind_conf->backlog)
+		return l->bind_conf->backlog;
 
 	if (l->bind_conf->frontend->backlog)
 		return l->bind_conf->frontend->backlog;
 
-	if (l->maxconn)
-		return l->maxconn;
+	if (l->bind_conf->maxconn)
+		return l->bind_conf->maxconn;
 
 	if (l->bind_conf->frontend->maxconn)
 		return l->bind_conf->frontend->maxconn;
@@ -919,12 +1001,12 @@ void listener_accept(struct listener *l)
 
 	p = l->bind_conf->frontend;
 
-	/* if l->maxaccept is -1, then max_accept is UINT_MAX. It is not really
-	 * illimited, but it is probably enough.
+	/* if l->bind_conf->maxaccept is -1, then max_accept is UINT_MAX. It is
+	 * not really illimited, but it is probably enough.
 	 */
-	max_accept = l->maxaccept ? l->maxaccept : 1;
+	max_accept = l->bind_conf->maxaccept ? l->bind_conf->maxaccept : 1;
 
-	if (!(l->options & LI_O_UNLIMITED) && global.sps_lim) {
+	if (!(l->bind_conf->options & BC_O_UNLIMITED) && global.sps_lim) {
 		int max = freq_ctr_remain(&global.sess_per_sec, global.sps_lim, 0);
 
 		if (unlikely(!max)) {
@@ -937,7 +1019,7 @@ void listener_accept(struct listener *l)
 			max_accept = max;
 	}
 
-	if (!(l->options & LI_O_UNLIMITED) && global.cps_lim) {
+	if (!(l->bind_conf->options & BC_O_UNLIMITED) && global.cps_lim) {
 		int max = freq_ctr_remain(&global.conn_per_sec, global.cps_lim, 0);
 
 		if (unlikely(!max)) {
@@ -950,7 +1032,7 @@ void listener_accept(struct listener *l)
 			max_accept = max;
 	}
 #ifdef USE_OPENSSL
-	if (!(l->options & LI_O_UNLIMITED) && global.ssl_lim &&
+	if (!(l->bind_conf->options & BC_O_UNLIMITED) && global.ssl_lim &&
 	    l->bind_conf && l->bind_conf->options & BC_O_USE_SSL) {
 		int max = freq_ctr_remain(&global.ssl_per_sec, global.ssl_lim, 0);
 
@@ -993,7 +1075,7 @@ void listener_accept(struct listener *l)
 		 */
 		do {
 			count = l->nbconn;
-			if (unlikely(l->maxconn && count >= l->maxconn)) {
+			if (unlikely(l->bind_conf->maxconn && count >= l->bind_conf->maxconn)) {
 				/* the listener was marked full or another
 				 * thread is going to do it.
 				 */
@@ -1019,7 +1101,7 @@ void listener_accept(struct listener *l)
 			} while (!_HA_ATOMIC_CAS(&p->feconn, &count, next_feconn));
 		}
 
-		if (!(l->options & LI_O_UNLIMITED)) {
+		if (!(l->bind_conf->options & BC_O_UNLIMITED)) {
 			do {
 				count = actconn;
 				if (unlikely(count >= global.maxconn)) {
@@ -1055,7 +1137,7 @@ void listener_accept(struct listener *l)
 				_HA_ATOMIC_DEC(&l->nbconn);
 				if (p)
 					_HA_ATOMIC_DEC(&p->feconn);
-				if (!(l->options & LI_O_UNLIMITED))
+				if (!(l->bind_conf->options & BC_O_UNLIMITED))
 					_HA_ATOMIC_DEC(&actconn);
 				continue;
 
@@ -1077,14 +1159,20 @@ void listener_accept(struct listener *l)
 			proxy_inc_fe_conn_ctr(l, p);
 		}
 
-		if (!(l->options & LI_O_UNLIMITED)) {
+		if (!(l->bind_conf->options & BC_O_UNLIMITED)) {
 			count = update_freq_ctr(&global.conn_per_sec, 1);
 			HA_ATOMIC_UPDATE_MAX(&global.cps_max, count);
 		}
 
 		_HA_ATOMIC_INC(&activity[tid].accepted);
 
-		/* past this point, l->accept() will automatically decrement
+		/* count the number of times an accepted connection resulted in
+		 * maxconn being reached.
+		 */
+		if (unlikely(_HA_ATOMIC_LOAD(&actconn) + 1 >= global.maxconn))
+			_HA_ATOMIC_INC(&maxconn_reached);
+
+		/* past this point, l->bind_conf->accept() will automatically decrement
 		 * l->nbconn, feconn and actconn once done. Setting next_*conn=0
 		 * allows the error path not to rollback on nbconn. It's more
 		 * convenient than duplicating all exit labels.
@@ -1095,64 +1183,183 @@ void listener_accept(struct listener *l)
 
 
 #if defined(USE_THREAD)
+		if (!(global.tune.options & GTUNE_LISTENER_MQ_ANY) || stopping)
+			goto local_accept;
+
+		/* we want to perform thread rebalancing if the listener is
+		 * bound to more than one thread or if it's part of a shard
+		 * with more than one listener.
+		 */
 		mask = l->rx.bind_thread & _HA_ATOMIC_LOAD(&tg->threads_enabled);
-		if (atleast2(mask) && (global.tune.options & GTUNE_LISTENER_MQ) && !stopping) {
+		if (l->rx.shard_info || atleast2(mask)) {
 			struct accept_queue_ring *ring;
-			unsigned int t, t0, t1, t2;
-			int base = tg->base;
+			struct listener *new_li;
+			uint r1, r2, t, t1, t2;
+			ulong n0, n1;
+			const struct tgroup_info *g1, *g2;
+			ulong m1, m2;
+			ulong *thr_idx_ptr;
 
 			/* The principle is that we have two running indexes,
 			 * each visiting in turn all threads bound to this
-			 * listener. The connection will be assigned to the one
-			 * with the least connections, and the other one will
-			 * be updated. This provides a good fairness on short
-			 * connections (round robin) and on long ones (conn
-			 * count), without ever missing any idle thread.
+			 * listener's shard. The connection will be assigned to
+			 * the one with the least connections, and the other
+			 * one will be updated. This provides a good fairness
+			 * on short connections (round robin) and on long ones
+			 * (conn count), without ever missing any idle thread.
+			 * Each thread number is encoded as a combination of
+			 * times the receiver number and its local thread
+			 * number from 0 to MAX_THREADS_PER_GROUP - 1. The two
+			 * indexes are stored as 10/12 bit numbers in the thr_idx
+			 * array, since there are up to LONGBITS threads and
+			 * groups that can be represented. They are represented
+			 * like this:
+			 *           31:20    19:15   14:10    9:5     4:0
+			 *   32b: [ counter | r2num | t2num | r1num | t1num ]
+			 *
+			 *           63:24    23:18   17:12   11:6     5:0
+			 *   64b: [ counter | r2num | t2num | r1num | t1num ]
+			 *
+			 * The change counter is only used to avoid swapping too
+			 * old a value when the value loops back.
+			 *
+			 * In the loop below we have this for each index:
+			 *   - n is the thread index
+			 *   - r is the receiver number
+			 *   - g is the receiver's thread group
+			 *   - t is the thread number in this receiver
+			 *   - m is the receiver's thread mask shifted by the thread number
 			 */
 
 			/* keep a copy for the final update. thr_idx is composite
-			 * and made of (t2<<16) + t1.
+			 * and made of (n2<<16) + n1.
 			 */
-			t0 = l->thr_idx;
-			do {
-				unsigned long m1, m2;
-				int q1, q2;
+			thr_idx_ptr = l->rx.shard_info ? &((struct listener *)(l->rx.shard_info->ref->owner))->thr_idx : &l->thr_idx;
+			while (1) {
+				int q0, q1, q2;
 
-				t2 = t1 = t0;
-				t2 >>= 16;
-				t1 &= 0xFFFF;
+				/* calculate r1/g1/t1 first (ascending idx) */
+				n0 = _HA_ATOMIC_LOAD(thr_idx_ptr);
+				new_li = NULL;
 
-				/* t1 walks low to high bits ;
-				 * t2 walks high to low.
-				 */
-				m1 = mask >> t1;
-				m2 = mask & (t2 ? nbits(t2 + 1) : ~0UL);
+				t1 = (uint)n0 & (LONGBITS - 1);
+				r1 = ((uint)n0 / LONGBITS) & (LONGBITS - 1);
 
-				if (unlikely(!(m1 & 1))) {
-					m1 &= ~1UL;
-					if (!m1) {
-						m1 = mask;
-						t1 = 0;
+				while (1) {
+					if (l->rx.shard_info) {
+						/* multiple listeners, take the group into account */
+						if (r1 >= l->rx.shard_info->nbgroups)
+							r1 = 0;
+
+						g1 = &ha_tgroup_info[l->rx.shard_info->members[r1]->bind_tgroup - 1];
+						m1 = l->rx.shard_info->members[r1]->bind_thread;
+					} else {
+						/* single listener */
+						r1 = 0;
+						g1 = tg;
+						m1 = l->rx.bind_thread;
 					}
-					t1 += my_ffsl(m1) - 1;
+					m1 &= _HA_ATOMIC_LOAD(&g1->threads_enabled);
+					m1 >>= t1;
+
+					/* find first existing thread */
+					if (unlikely(!(m1 & 1))) {
+						m1 &= ~1UL;
+						if (!m1) {
+							/* no more threads here, switch to
+							 * first thread of next group.
+							 */
+							t1 = 0;
+							if (l->rx.shard_info)
+								r1++;
+							/* loop again */
+							continue;
+						}
+						t1 += my_ffsl(m1) - 1;
+					}
+					/* done: r1 and t1 are OK */
+					break;
 				}
 
-				if (unlikely(!(m2 & (1UL << t2)) || t1 == t2)) {
-					/* highest bit not set */
-					if (!m2)
-						m2 = mask;
+				/* now r2/g2/t2 (descending idx) */
+				t2 = ((uint)n0 / LONGBITS / LONGBITS) & (LONGBITS - 1);
+				r2 = ((uint)n0 / LONGBITS / LONGBITS / LONGBITS) & (LONGBITS - 1);
 
-					t2 = my_flsl(m2) - 1;
+				/* if running in round-robin mode ("fair"), we don't need
+				 * to go further.
+				 */
+				if ((global.tune.options & GTUNE_LISTENER_MQ_ANY) == GTUNE_LISTENER_MQ_FAIR) {
+					t = g1->base + t1;
+					if (l->rx.shard_info && t != tid)
+						new_li = l->rx.shard_info->members[r1]->owner;
+					goto updt_t1;
 				}
 
-				/* now we have two distinct thread IDs belonging to the mask */
-				q1 = accept_queue_rings[base + t1].tail - accept_queue_rings[base + t1].head + ACCEPT_QUEUE_SIZE;
-				if (q1 >= ACCEPT_QUEUE_SIZE)
-					q1 -= ACCEPT_QUEUE_SIZE;
+				while (1) {
+					if (l->rx.shard_info) {
+						/* multiple listeners, take the group into account */
+						if (r2 >= l->rx.shard_info->nbgroups)
+							r2 = l->rx.shard_info->nbgroups - 1;
 
-				q2 = accept_queue_rings[base + t2].tail - accept_queue_rings[base + t2].head + ACCEPT_QUEUE_SIZE;
-				if (q2 >= ACCEPT_QUEUE_SIZE)
-					q2 -= ACCEPT_QUEUE_SIZE;
+						g2 = &ha_tgroup_info[l->rx.shard_info->members[r2]->bind_tgroup - 1];
+						m2 = l->rx.shard_info->members[r2]->bind_thread;
+					} else {
+						/* single listener */
+						r2 = 0;
+						g2 = tg;
+						m2 = l->rx.bind_thread;
+					}
+					m2 &= _HA_ATOMIC_LOAD(&g2->threads_enabled);
+					m2 &= nbits(t2 + 1);
+
+					/* find previous existing thread */
+					if (unlikely(!(m2 & (1UL << t2)) || (g1 == g2 && t1 == t2))) {
+						/* highest bit not set or colliding threads, let's check
+						 * if we still have other threads available after this
+						 * one.
+						 */
+						m2 &= ~(1UL << t2);
+						if (!m2) {
+							/* no more threads here, switch to
+							 * last thread of previous group.
+							 */
+							t2 = MAX_THREADS_PER_GROUP - 1;
+							if (l->rx.shard_info)
+								r2--;
+							/* loop again */
+							continue;
+						}
+						t2 = my_flsl(m2) - 1;
+					}
+					/* done: r2 and t2 are OK */
+					break;
+				}
+
+				/* tests show that it's worth checking that other threads have not
+				 * already changed the index to save the rest of the calculation,
+				 * or we'd have to redo it anyway.
+				 */
+				if (n0 != _HA_ATOMIC_LOAD(thr_idx_ptr))
+					continue;
+
+				/* here we have (r1,g1,t1) that designate the first receiver, its
+				 * thread group and local thread, and (r2,g2,t2) that designate
+				 * the second receiver, its thread group and local thread. We'll
+				 * also consider the local thread with q0.
+				 */
+				q0 = accept_queue_ring_len(&accept_queue_rings[tid]);
+				q1 = accept_queue_ring_len(&accept_queue_rings[g1->base + t1]);
+				q2 = accept_queue_ring_len(&accept_queue_rings[g2->base + t2]);
+
+				/* add to this the currently active connections */
+				q0 += _HA_ATOMIC_LOAD(&l->thr_conn[ti->ltid]);
+				if (l->rx.shard_info) {
+					q1 += _HA_ATOMIC_LOAD(&((struct listener *)l->rx.shard_info->members[r1]->owner)->thr_conn[t1]);
+					q2 += _HA_ATOMIC_LOAD(&((struct listener *)l->rx.shard_info->members[r2]->owner)->thr_conn[t2]);
+				} else {
+					q1 += _HA_ATOMIC_LOAD(&l->thr_conn[t1]);
+					q2 += _HA_ATOMIC_LOAD(&l->thr_conn[t2]);
+				}
 
 				/* we have 3 possibilities now :
 				 *   q1 < q2 : t1 is less loaded than t2, so we pick it
@@ -1164,34 +1371,79 @@ void listener_accept(struct listener *l)
 				 *   q1 = q2 : both are equally loaded, thus we pick t1
 				 *             and update t1 as it will become more loaded
 				 *             than t2.
+				 * On top of that, if in the end the current thread appears
+				 * to be as good of a deal, we'll prefer it over a foreign
+				 * one as it will improve locality and avoid a migration.
 				 */
 
-				q1 += l->thr_conn[t1];
-				q2 += l->thr_conn[t2];
-
 				if (q1 - q2 < 0) {
-					t = t1;
-					t2 = t2 ? t2 - 1 : LONGBITS - 1;
+					t = g1->base + t1;
+					if (q0 <= q1)
+						t = tid;
+
+					if (l->rx.shard_info && t != tid)
+						new_li = l->rx.shard_info->members[r1]->owner;
+
+					t2--;
+					if (t2 >= MAX_THREADS_PER_GROUP) {
+						if (l->rx.shard_info)
+							r2--;
+						t2 = MAX_THREADS_PER_GROUP - 1;
+					}
 				}
 				else if (q1 - q2 > 0) {
-					t = t2;
-					t1++;
-					if (t1 >= LONGBITS)
-						t1 = 0;
+					t = g2->base + t2;
+					if (q0 <= q2)
+						t = tid;
+
+					if (l->rx.shard_info && t != tid)
+						new_li = l->rx.shard_info->members[r2]->owner;
+					goto updt_t1;
 				}
-				else {
-					t = t1;
+				else { // q1 == q2
+					t = g1->base + t1;
+					if (q0 < q1) // local must be strictly better than both
+						t = tid;
+
+					if (l->rx.shard_info && t != tid)
+						new_li = l->rx.shard_info->members[r1]->owner;
+				updt_t1:
 					t1++;
-					if (t1 >= LONGBITS)
+					if (t1 >= MAX_THREADS_PER_GROUP) {
+						if (l->rx.shard_info)
+							r1++;
 						t1 = 0;
+					}
 				}
 
-				/* new value for thr_idx */
-				t1 += (t2 << 16);
-			} while (unlikely(!_HA_ATOMIC_CAS(&l->thr_idx, &t0, t1)));
+				/* The target thread number is in <t> now. Let's
+				 * compute the new index and try to update it.
+				 */
+
+				/* take previous counter and increment it */
+				n1  = n0 & -(ulong)(LONGBITS * LONGBITS * LONGBITS * LONGBITS);
+				n1 += LONGBITS * LONGBITS * LONGBITS * LONGBITS;
+				n1 += (((r2 * LONGBITS) + t2) * LONGBITS * LONGBITS);
+				n1 += (r1 * LONGBITS) + t1;
+				if (likely(_HA_ATOMIC_CAS(thr_idx_ptr, &n0, n1)))
+					break;
+
+				/* bah we lost the race, try again */
+				__ha_cpu_relax();
+			} /* end of main while() loop */
+
+			/* we may need to update the listener in the connection
+			 * if we switched to another group.
+			 */
+			if (new_li)
+				cli_conn->target = &new_li->obj_type;
+
+			/* here we have the target thread number in <t> and we hold a
+			 * reservation in the target ring.
+			 */
 
 			if (l->rx.proto && l->rx.proto->set_affinity) {
-				if (l->rx.proto->set_affinity(cli_conn, base + t)) {
+				if (l->rx.proto->set_affinity(cli_conn, t)) {
 					/* Failed migration, stay on the same thread. */
 					goto local_accept;
 				}
@@ -1203,22 +1455,24 @@ void listener_accept(struct listener *l)
 			 * performing model, likely due to better cache locality
 			 * when processing this loop.
 			 */
-			ring = &accept_queue_rings[base + t];
+			ring = &accept_queue_rings[t];
 			if (accept_queue_push_mp(ring, cli_conn)) {
-				_HA_ATOMIC_INC(&activity[base + t].accq_pushed);
+				_HA_ATOMIC_INC(&activity[t].accq_pushed);
 				tasklet_wakeup(ring->tasklet);
 				continue;
 			}
 			/* If the ring is full we do a synchronous accept on
 			 * the local thread here.
 			 */
-			_HA_ATOMIC_INC(&activity[base + t].accq_full);
+			_HA_ATOMIC_INC(&activity[t].accq_full);
 		}
 #endif // USE_THREAD
 
  local_accept:
+		/* restore the connection's listener in case we failed to migrate above */
+		cli_conn->target = &l->obj_type;
 		_HA_ATOMIC_INC(&l->thr_conn[ti->ltid]);
-		ret = l->accept(cli_conn);
+		ret = l->bind_conf->accept(cli_conn);
 		if (unlikely(ret <= 0)) {
 			/* The connection was closed by stream_accept(). Either
 			 * we just have to ignore it (ret == 0) or it's a critical
@@ -1232,14 +1486,15 @@ void listener_accept(struct listener *l)
 		}
 
 		/* increase the per-process number of cumulated sessions, this
-		 * may only be done once l->accept() has accepted the connection.
+		 * may only be done once l->bind_conf->accept() has accepted the
+		 * connection.
 		 */
-		if (!(l->options & LI_O_UNLIMITED)) {
+		if (!(l->bind_conf->options & BC_O_UNLIMITED)) {
 			count = update_freq_ctr(&global.sess_per_sec, 1);
 			HA_ATOMIC_UPDATE_MAX(&global.sps_max, count);
 		}
 #ifdef USE_OPENSSL
-		if (!(l->options & LI_O_UNLIMITED) &&
+		if (!(l->bind_conf->options & BC_O_UNLIMITED) &&
 		    l->bind_conf && l->bind_conf->options & BC_O_USE_SSL) {
 			count = update_freq_ctr(&global.ssl_per_sec, 1);
 			HA_ATOMIC_UPDATE_MAX(&global.ssl_max, count);
@@ -1259,7 +1514,7 @@ void listener_accept(struct listener *l)
 	if (next_actconn)
 		_HA_ATOMIC_DEC(&actconn);
 
-	if ((l->state == LI_FULL && (!l->maxconn || l->nbconn < l->maxconn)) ||
+	if ((l->state == LI_FULL && (!l->bind_conf->maxconn || l->nbconn < l->bind_conf->maxconn)) ||
 	    (l->state == LI_LIMITED &&
 	     ((!p || p->feconn < p->maxconn) && (actconn < global.maxconn) &&
 	      (!tick_isset(global_listener_queue_task->expire) ||
@@ -1316,7 +1571,7 @@ void listener_release(struct listener *l)
 {
 	struct proxy *fe = l->bind_conf->frontend;
 
-	if (!(l->options & LI_O_UNLIMITED))
+	if (!(l->bind_conf->options & BC_O_UNLIMITED))
 		_HA_ATOMIC_DEC(&actconn);
 	if (fe)
 		_HA_ATOMIC_DEC(&fe->feconn);
@@ -1389,6 +1644,163 @@ struct task *manage_global_listener_queue(struct task *t, void *context, unsigne
 	return t;
 }
 
+/* Applies the thread mask, shards etc to the bind_conf. It normally returns 0
+ * otherwie the number of errors. Upon error it may set error codes (ERR_*) in
+ * err_code. It is supposed to be called only once very late in the boot process
+ * after the bind_conf's thread_set is fixed. The function may emit warnings and
+ * alerts. Extra listeners may be created on the fly.
+ */
+int bind_complete_thread_setup(struct bind_conf *bind_conf, int *err_code)
+{
+	struct proxy *fe = bind_conf->frontend;
+	struct listener *li, *new_li, *ref;
+	struct thread_set new_ts;
+	int shard, shards, todo, done, grp, dups;
+	ulong mask, gmask, bit;
+	int cfgerr = 0;
+	char *err;
+
+	err = NULL;
+	if (thread_resolve_group_mask(&bind_conf->thread_set, 0, &err) < 0) {
+		ha_alert("%s '%s': %s in 'bind %s' at [%s:%d].\n",
+			 proxy_type_str(fe),
+			 fe->id, err, bind_conf->arg, bind_conf->file, bind_conf->line);
+		free(err);
+		cfgerr++;
+		return cfgerr;
+	}
+
+	/* apply thread masks and groups to all receivers */
+	list_for_each_entry(li, &bind_conf->listeners, by_bind) {
+		shards = bind_conf->settings.shards;
+		todo = thread_set_count(&bind_conf->thread_set);
+
+		/* special values: -1 = "by-thread", -2 = "by-group" */
+		if (shards == -1) {
+			if (protocol_supports_flag(li->rx.proto, PROTO_F_REUSEPORT_SUPPORTED))
+				shards = todo;
+			else {
+				if (fe != global.cli_fe)
+					ha_diag_warning("[%s:%d]: Disabling per-thread sharding for listener in"
+					                " %s '%s' because SO_REUSEPORT is disabled\n",
+					                bind_conf->file, bind_conf->line, proxy_type_str(fe), fe->id);
+				shards = 1;
+			}
+		}
+		else if (shards == -2)
+			shards = protocol_supports_flag(li->rx.proto, PROTO_F_REUSEPORT_SUPPORTED) ? my_popcountl(bind_conf->thread_set.grps) : 1;
+
+		/* no more shards than total threads */
+		if (shards > todo)
+			shards = todo;
+
+		/* We also need to check if an explicit shards count was set and cannot be honored */
+		if (shards > 1 && !protocol_supports_flag(li->rx.proto, PROTO_F_REUSEPORT_SUPPORTED)) {
+			ha_warning("[%s:%d]: Disabling sharding for listener in %s '%s' because SO_REUSEPORT is disabled\n",
+			           bind_conf->file, bind_conf->line, proxy_type_str(fe), fe->id);
+			shards = 1;
+		}
+
+		shard = done = grp = bit = mask = 0;
+		new_li = li;
+
+		while (shard < shards) {
+			memset(&new_ts, 0, sizeof(new_ts));
+			while (grp < global.nbtgroups && done < todo) {
+				/* enlarge mask to cover next bit of bind_thread till we
+				 * have enough bits for one shard. We restart from the
+				 * current grp+bit.
+				 */
+
+				/* first let's find the first non-empty group starting at <mask> */
+				if (!(bind_conf->thread_set.rel[grp] & ha_tgroup_info[grp].threads_enabled & ~mask)) {
+					grp++;
+					mask = 0;
+					continue;
+				}
+
+				/* take next unassigned bit */
+				bit = (bind_conf->thread_set.rel[grp] & ~mask) & -(bind_conf->thread_set.rel[grp] & ~mask);
+				new_ts.rel[grp] |= bit;
+				mask |= bit;
+				new_ts.grps |= 1UL << grp;
+
+				done += shards;
+			};
+
+			BUG_ON(!new_ts.grps); // no more bits left unassigned
+
+			/* Create all required listeners for all bound groups. If more than one group is
+			 * needed, the first receiver serves as a reference, and subsequent ones point to
+			 * it. We already have a listener available in new_li() so we only allocate a new
+			 * one if we're not on the last one. We count the remaining groups by copying their
+			 * mask into <gmask> and dropping the lowest bit at the end of the loop until there
+			 * is no more. Ah yes, it's not pretty :-/
+			 */
+			ref = new_li;
+			gmask = new_ts.grps;
+			for (dups = 0; gmask; dups++) {
+				/* assign the first (and only) thread and group */
+				new_li->rx.bind_thread = thread_set_nth_tmask(&new_ts, dups);
+				new_li->rx.bind_tgroup = thread_set_nth_group(&new_ts, dups);
+
+				if (dups) {
+					/* it has been allocated already in the previous round */
+					shard_info_attach(&new_li->rx, ref->rx.shard_info);
+					new_li->rx.flags |= RX_F_MUST_DUP;
+				}
+
+				gmask &= gmask - 1; // drop lowest bit
+				if (gmask) {
+					/* yet another listener expected in this shard, let's
+					 * chain it.
+					 */
+					struct listener *tmp_li = clone_listener(new_li);
+
+					if (!tmp_li) {
+						ha_alert("Out of memory while trying to allocate extra listener for group %u of shard %d in %s %s\n",
+							 new_li->rx.bind_tgroup, shard, proxy_type_str(fe), fe->id);
+						cfgerr++;
+						*err_code |= ERR_FATAL | ERR_ALERT;
+						return cfgerr;
+					}
+
+					/* if we're forced to create at least two listeners, we have to
+					 * allocate a shared shard_info that's linked to from the reference
+					 * and each other listener, so we'll create it here.
+					 */
+					if (!shard_info_attach(&ref->rx, NULL)) {
+						ha_alert("Out of memory while trying to allocate shard_info for listener for group %u of shard %d in %s %s\n",
+							 new_li->rx.bind_tgroup, shard, proxy_type_str(fe), fe->id);
+						cfgerr++;
+						*err_code |= ERR_FATAL | ERR_ALERT;
+						return cfgerr;
+					}
+					new_li = tmp_li;
+				}
+			}
+			done -= todo;
+
+			shard++;
+			if (shard >= shards)
+				break;
+
+			/* create another listener for new shards */
+			new_li = clone_listener(li);
+			if (!new_li) {
+				ha_alert("Out of memory while trying to allocate extra listener for shard %d in %s %s\n",
+					 shard, proxy_type_str(fe), fe->id);
+				cfgerr++;
+				*err_code |= ERR_FATAL | ERR_ALERT;
+				return cfgerr;
+			}
+		}
+	}
+
+	/* success */
+	return cfgerr;
+}
+
 /*
  * Registers the bind keyword list <kwl> as a list of valid keywords for next
  * parsing sessions.
@@ -1514,9 +1926,10 @@ struct bind_conf *bind_conf_alloc(struct proxy *fe, const char *file,
 	bind_conf->settings.ux.uid = -1;
 	bind_conf->settings.ux.gid = -1;
 	bind_conf->settings.ux.mode = 0;
-	bind_conf->settings.shards = 1;
+	bind_conf->settings.shards = global.tune.default_shards;
 	bind_conf->xprt = xprt;
 	bind_conf->frontend = fe;
+	bind_conf->analysers = fe->fe_req_ana;
 	bind_conf->severity_output = CLI_SEVERITY_NONE;
 #ifdef USE_OPENSSL
 	HA_RWLOCK_INIT(&bind_conf->sni_lock);
@@ -1584,18 +1997,13 @@ smp_fetch_so_name(const struct arg *args, struct sample *smp, const char *kw, vo
 /* parse the "accept-proxy" bind keyword */
 static int bind_parse_accept_proxy(char **args, int cur_arg, struct proxy *px, struct bind_conf *conf, char **err)
 {
-	struct listener *l;
-
-	list_for_each_entry(l, &conf->listeners, by_bind)
-		l->options |= LI_O_ACC_PROXY;
-
+	conf->options |= BC_O_ACC_PROXY;
 	return 0;
 }
 
 /* parse the "accept-netscaler-cip" bind keyword */
 static int bind_parse_accept_netscaler_cip(char **args, int cur_arg, struct proxy *px, struct bind_conf *conf, char **err)
 {
-	struct listener *l;
 	uint32_t val;
 
 	if (!*args[cur_arg + 1]) {
@@ -1609,18 +2017,14 @@ static int bind_parse_accept_netscaler_cip(char **args, int cur_arg, struct prox
 		return ERR_ALERT | ERR_FATAL;
 	}
 
-	list_for_each_entry(l, &conf->listeners, by_bind) {
-		l->options |= LI_O_ACC_CIP;
-		conf->ns_cip_magic = val;
-	}
-
+	conf->options |= BC_O_ACC_CIP;
+	conf->ns_cip_magic = val;
 	return 0;
 }
 
 /* parse the "backlog" bind keyword */
 static int bind_parse_backlog(char **args, int cur_arg, struct proxy *px, struct bind_conf *conf, char **err)
 {
-	struct listener *l;
 	int val;
 
 	if (!*args[cur_arg + 1]) {
@@ -1634,9 +2038,7 @@ static int bind_parse_backlog(char **args, int cur_arg, struct proxy *px, struct
 		return ERR_ALERT | ERR_FATAL;
 	}
 
-	list_for_each_entry(l, &conf->listeners, by_bind)
-		l->backlog = val;
-
+	conf->backlog = val;
 	return 0;
 }
 
@@ -1787,7 +2189,6 @@ int bind_parse_args_list(struct bind_conf *bind_conf, char **args, int cur_arg,
 /* parse the "maxconn" bind keyword */
 static int bind_parse_maxconn(char **args, int cur_arg, struct proxy *px, struct bind_conf *conf, char **err)
 {
-	struct listener *l;
 	int val;
 
 	if (!*args[cur_arg + 1]) {
@@ -1801,9 +2202,7 @@ static int bind_parse_maxconn(char **args, int cur_arg, struct proxy *px, struct
 		return ERR_ALERT | ERR_FATAL;
 	}
 
-	list_for_each_entry(l, &conf->listeners, by_bind)
-		l->maxconn = val;
-
+	conf->maxconn = val;
 	return 0;
 }
 
@@ -1826,7 +2225,6 @@ static int bind_parse_name(char **args, int cur_arg, struct proxy *px, struct bi
 /* parse the "nice" bind keyword */
 static int bind_parse_nice(char **args, int cur_arg, struct proxy *px, struct bind_conf *conf, char **err)
 {
-	struct listener *l;
 	int val;
 
 	if (!*args[cur_arg + 1]) {
@@ -1840,9 +2238,7 @@ static int bind_parse_nice(char **args, int cur_arg, struct proxy *px, struct bi
 		return ERR_ALERT | ERR_FATAL;
 	}
 
-	list_for_each_entry(l, &conf->listeners, by_bind)
-		l->nice = val;
-
+	conf->nice = val;
 	return 0;
 }
 
@@ -1872,7 +2268,7 @@ static int bind_parse_proto(char **args, int cur_arg, struct proxy *px, struct b
 	return 0;
 }
 
-/* parse the "shards" bind keyword. Takes an integer or "by-thread" */
+/* parse the "shards" bind keyword. Takes an integer, "by-thread", or "by-group" */
 static int bind_parse_shards(char **args, int cur_arg, struct proxy *px, struct bind_conf *conf, char **err)
 {
 	int val;
@@ -1883,7 +2279,9 @@ static int bind_parse_shards(char **args, int cur_arg, struct proxy *px, struct
 	}
 
 	if (strcmp(args[cur_arg + 1], "by-thread") == 0) {
-		val = MAX_THREADS; /* will be trimmed later anyway */
+		val = -1; /* -1 = "by-thread", will be fixed in check_config_validity() */
+	} else if (strcmp(args[cur_arg + 1], "by-group") == 0) {
+		val = -2; /* -2 = "by-group", will be fixed in check_config_validity() */
 	} else {
 		val = atol(args[cur_arg + 1]);
 		if (val < 1 || val > MAX_THREADS) {
@@ -1896,45 +2294,40 @@ static int bind_parse_shards(char **args, int cur_arg, struct proxy *px, struct
 	return 0;
 }
 
-/* parse the "thread" bind keyword */
+/* parse the "thread" bind keyword. This will replace any preset thread_set */
 static int bind_parse_thread(char **args, int cur_arg, struct proxy *px, struct bind_conf *conf, char **err)
 {
-	char *sep = NULL;
-	ulong thread = 0;
-	long tgroup = 0;
-
-	tgroup = strtol(args[cur_arg + 1], &sep, 10);
-	if (*sep == '/') {
-		/* a thread group was present */
-		if (tgroup < 1 || tgroup > MAX_TGROUPS) {
-			memprintf(err, "'%s' thread-group number must be between 1 and %d (was %ld)", args[cur_arg + 1], MAX_TGROUPS, tgroup);
-			return ERR_ALERT | ERR_FATAL;
-		}
-		sep++;
-	}
-	else {
-		/* no thread group */
-		tgroup = 0;
-		sep = args[cur_arg + 1];
-	}
-
-	if ((conf->bind_tgroup || conf->bind_thread) &&
-	    conf->bind_tgroup != tgroup) {
-		memprintf(err, "'%s' multiple thread-groups are not supported", args[cur_arg + 1]);
+	/* note that the thread set is zeroed before first call, and we don't
+	 * want to reset it so that it remains possible to chain multiple
+	 * "thread" directives.
+	 */
+	if (parse_thread_set(args[cur_arg+1], &conf->thread_set, err) < 0)
 		return ERR_ALERT | ERR_FATAL;
-	}
-
-	if (parse_process_number(sep, &thread, LONGBITS, NULL, err)) {
-		memprintf(err, "'%s' : %s", sep, *err);
-		return ERR_ALERT | ERR_FATAL;
-	}
-
-	conf->bind_thread |= thread;
-	conf->bind_tgroup  = tgroup;
 	return 0;
 }
 
-/* config parser for global "tune.listener.multi-queue", accepts "on" or "off" */
+/* config parser for global "tune.listener.default-shards" */
+static int cfg_parse_tune_listener_shards(char **args, int section_type, struct proxy *curpx,
+                                          const struct proxy *defpx, const char *file, int line,
+                                          char **err)
+{
+	if (too_many_args(1, args, err, NULL))
+		return -1;
+
+	if (strcmp(args[1], "by-thread") == 0)
+		global.tune.default_shards = -1;
+	else if (strcmp(args[1], "by-group") == 0)
+		global.tune.default_shards = -2;
+	else if (strcmp(args[1], "by-process") == 0)
+		global.tune.default_shards = 1;
+	else {
+		memprintf(err, "'%s' expects either 'by-process', 'by-group', or 'by-thread' but got '%s'.", args[0], args[1]);
+		return -1;
+	}
+	return 0;
+}
+
+/* config parser for global "tune.listener.multi-queue", accepts "on", "fair" or "off" */
 static int cfg_parse_tune_listener_mq(char **args, int section_type, struct proxy *curpx,
                                       const struct proxy *defpx, const char *file, int line,
                                       char **err)
@@ -1943,11 +2336,13 @@ static int cfg_parse_tune_listener_mq(char **args, int section_type, struct prox
 		return -1;
 
 	if (strcmp(args[1], "on") == 0)
-		global.tune.options |= GTUNE_LISTENER_MQ;
+		global.tune.options = (global.tune.options & ~GTUNE_LISTENER_MQ_ANY) | GTUNE_LISTENER_MQ_OPT;
+	else if (strcmp(args[1], "fair") == 0)
+		global.tune.options = (global.tune.options & ~GTUNE_LISTENER_MQ_ANY) | GTUNE_LISTENER_MQ_FAIR;
 	else if (strcmp(args[1], "off") == 0)
-		global.tune.options &= ~GTUNE_LISTENER_MQ;
+		global.tune.options &= ~GTUNE_LISTENER_MQ_ANY;
 	else {
-		memprintf(err, "'%s' expects either 'on' or 'off' but got '%s'.", args[0], args[1]);
+		memprintf(err, "'%s' expects either 'on', 'fair', or 'off' but got '%s'.", args[0], args[1]);
 		return -1;
 	}
 	return 0;
@@ -2000,6 +2395,7 @@ INITCALL1(STG_REGISTER, bind_register_keywords, &bind_kws);
 
 /* config keyword parsers */
 static struct cfg_kw_list cfg_kws = {ILH, {
+	{ CFG_GLOBAL, "tune.listener.default-shards",   cfg_parse_tune_listener_shards  },
 	{ CFG_GLOBAL, "tune.listener.multi-queue",      cfg_parse_tune_listener_mq      },
 	{ 0, NULL, NULL }
 }};
