@@ -20,7 +20,7 @@ void quic_loss_srtt_update(struct quic_loss *ql,
                            struct quic_conn *qc)
 {
 	TRACE_ENTER(QUIC_EV_CONN_RTTUPDT, qc);
-	TRACE_DEVEL("Loss info update", QUIC_EV_CONN_RTTUPDT, qc, &rtt, &ack_delay, ql);
+	TRACE_PROTO("TX loss srtt update", QUIC_EV_CONN_RTTUPDT, qc, &rtt, &ack_delay, ql);
 
 	ql->latest_rtt = rtt;
 	if (!ql->rtt_min) {
@@ -37,7 +37,7 @@ void quic_loss_srtt_update(struct quic_loss *ql,
 		/* Specific to QUIC (RTT adjustment). */
 		if (ack_delay && rtt > ql->rtt_min + ack_delay)
 			rtt -= ack_delay;
-		diff = ql->srtt - rtt;
+		diff = (ql->srtt >> 3) - rtt;
 		if (diff < 0)
 			diff = -diff;
 		/* 4*rttvar = 3*rttvar + |diff| */
@@ -46,7 +46,7 @@ void quic_loss_srtt_update(struct quic_loss *ql,
 		ql->srtt += rtt - (ql->srtt >> 3);
 	}
 
-	TRACE_DEVEL("Loss info update", QUIC_EV_CONN_RTTUPDT, qc,,, ql);
+	TRACE_PROTO("TX loss srtt update", QUIC_EV_CONN_RTTUPDT, qc,,, ql);
 	TRACE_LEAVE(QUIC_EV_CONN_RTTUPDT, qc);
 }
 
@@ -62,11 +62,11 @@ struct quic_pktns *quic_loss_pktns(struct quic_conn *qc)
 	TRACE_ENTER(QUIC_EV_CONN_SPTO, qc);
 
 	pktns = &qc->pktns[QUIC_TLS_PKTNS_INITIAL];
-	TRACE_DEVEL("pktns", QUIC_EV_CONN_SPTO, qc, pktns);
+	TRACE_PROTO("TX loss pktns", QUIC_EV_CONN_SPTO, qc, pktns);
 	for (i = QUIC_TLS_PKTNS_HANDSHAKE; i < QUIC_TLS_PKTNS_MAX; i++) {
-		TRACE_DEVEL("pktns", QUIC_EV_CONN_SPTO, qc, &qc->pktns[i]);
+		TRACE_PROTO("TX loss pktns", QUIC_EV_CONN_SPTO, qc, &qc->pktns[i]);
 		if (!tick_isset(pktns->tx.loss_time) ||
-		    qc->pktns[i].tx.loss_time < pktns->tx.loss_time)
+		    tick_is_lt(qc->pktns[i].tx.loss_time, pktns->tx.loss_time))
 			pktns = &qc->pktns[i];
 	}
 
@@ -80,7 +80,7 @@ struct quic_pktns *quic_loss_pktns(struct quic_conn *qc)
  * as PTO value if not.
  */
 struct quic_pktns *quic_pto_pktns(struct quic_conn *qc,
-                                  int handshake_completed,
+                                  int handshake_confirmed,
                                   unsigned int *pto)
 {
 	int i;
@@ -93,19 +93,18 @@ struct quic_pktns *quic_pto_pktns(struct quic_conn *qc,
 		(ql->srtt >> 3) +
 		(QUIC_MAX(ql->rtt_var, QUIC_TIMER_GRANULARITY) << ql->pto_count);
 
-	if (!qc->path->in_flight) {
-		struct quic_enc_level *hel;
-
-		hel = &qc->els[QUIC_TLS_ENC_LEVEL_HANDSHAKE];
-		if (quic_tls_has_tx_sec(hel)) {
-			pktns = &qc->pktns[QUIC_TLS_PKTNS_HANDSHAKE];
-		}
-		else {
-			pktns = &qc->pktns[QUIC_TLS_PKTNS_INITIAL];
-		}
-		lpto = tick_add(now_ms, duration);
-		goto out;
-	}
+	/* RFC 9002 6.2.2.1. Before Address Validation
+	 *
+	 * the client MUST set the PTO timer if the client has not received an
+	 * acknowledgment for any of its Handshake packets and the handshake is
+	 * not confirmed (see Section 4.1.2 of [QUIC-TLS]), even if there are no
+	 * packets in flight.
+	 *
+	 * TODO implement the above paragraph for QUIC on backend side. Note
+	 * that if now_ms is used this function is not reentrant anymore and can
+	 * not be used anytime without side-effect (for example after QUIC
+	 * connection migration).
+	 */
 
 	lpto = TICK_ETERNITY;
 	pktns = p = &qc->pktns[QUIC_TLS_PKTNS_INITIAL];
@@ -117,8 +116,8 @@ struct quic_pktns *quic_pto_pktns(struct quic_conn *qc,
 			continue;
 
 		if (i == QUIC_TLS_PKTNS_01RTT) {
-			if (!handshake_completed) {
-				TRACE_STATE("handshake not already completed", QUIC_EV_CONN_SPTO, qc);
+			if (!handshake_confirmed) {
+				TRACE_STATE("TX PTO handshake not already confirmed", QUIC_EV_CONN_SPTO, qc);
 				pktns = p;
 				goto out;
 			}
@@ -128,17 +127,18 @@ struct quic_pktns *quic_pto_pktns(struct quic_conn *qc,
 
 		p = &qc->pktns[i];
 		tmp_pto = tick_add(p->tx.time_of_last_eliciting, duration);
-		if (!tick_isset(lpto) || tmp_pto < lpto) {
+		if (!tick_isset(lpto) || tick_is_lt(tmp_pto, lpto)) {
 			lpto = tmp_pto;
 			pktns = p;
 		}
-		TRACE_DEVEL("pktns", QUIC_EV_CONN_SPTO, qc, p);
+		TRACE_PROTO("TX PTO", QUIC_EV_CONN_SPTO, qc, p);
 	}
 
  out:
 	if (pto)
 		*pto = lpto;
-	TRACE_LEAVE(QUIC_EV_CONN_SPTO, qc, pktns, &duration);
+	TRACE_PROTO("TX PTO", QUIC_EV_CONN_SPTO, qc, pktns, &duration);
+	TRACE_LEAVE(QUIC_EV_CONN_SPTO, qc);
 
 	return pktns;
 }
@@ -159,7 +159,8 @@ void qc_packet_loss_lookup(struct quic_pktns *pktns, struct quic_conn *qc,
 	struct quic_loss *ql;
 	unsigned int loss_delay;
 
-	TRACE_ENTER(QUIC_EV_CONN_PKTLOSS, qc, pktns);
+	TRACE_ENTER(QUIC_EV_CONN_PKTLOSS, qc);
+	TRACE_PROTO("TX loss", QUIC_EV_CONN_PKTLOSS, qc, pktns);
 	pkts = &pktns->tx.pkts;
 	pktns->tx.loss_time = TICK_ETERNITY;
 	if (eb_is_empty(pkts))
@@ -188,7 +189,7 @@ void qc_packet_loss_lookup(struct quic_pktns *pktns, struct quic_conn *qc,
 			(int64_t)largest_acked_pn >= pkt->pn_node.key + QUIC_LOSS_PACKET_THRESHOLD) {
 			eb64_delete(&pkt->pn_node);
 			LIST_APPEND(lost_pkts, &pkt->list);
-			HA_ATOMIC_INC(&qc->prx_counters->lost_pkt);
+			ql->nb_lost_pkt++;
 		}
 		else {
 			if (tick_isset(pktns->tx.loss_time))
@@ -200,6 +201,7 @@ void qc_packet_loss_lookup(struct quic_pktns *pktns, struct quic_conn *qc,
 	}
 
  out:
-	TRACE_LEAVE(QUIC_EV_CONN_PKTLOSS, qc, pktns, lost_pkts);
+	TRACE_PROTO("TX loss", QUIC_EV_CONN_PKTLOSS, qc, pktns, lost_pkts);
+	TRACE_LEAVE(QUIC_EV_CONN_PKTLOSS, qc);
 }
 
