@@ -289,7 +289,7 @@ static int cli_parse_show_events(char **args, char *payload, struct appctx *appc
 /* Pre-configures a ring proxy to emit connections */
 void sink_setup_proxy(struct proxy *px)
 {
-	px->last_change = now.tv_sec;
+	px->last_change = ns_to_sec(now_ns);
 	px->cap = PR_CAP_BE;
 	px->maxconn = 0;
 	px->conn_retries = 1;
@@ -318,26 +318,13 @@ static void sink_forward_io_handler(struct appctx *appctx)
 	size_t len, cnt, ofs, last_ofs;
 	int ret = 0;
 
+	if (unlikely(se_fl_test(appctx->sedesc, (SE_FL_EOS|SE_FL_ERROR|SE_FL_SHR|SE_FL_SHW))))
+		goto out;
+
 	/* if stopping was requested, close immediately */
 	if (unlikely(stopping))
 		goto close;
 
-	/* for rex because it seems reset to timeout
-	 * and we don't want expire on this case
-	 * with a syslog server
-	 */
-	sc_oc(sc)->rex = TICK_ETERNITY;
-	/* rto should not change but it seems the case */
-	sc_oc(sc)->rto = TICK_ETERNITY;
-
-	/* an error was detected */
-	if (unlikely(sc_ic(sc)->flags & (CF_WRITE_ERROR|CF_SHUTW)))
-		goto close;
-
-	/* con closed by server side */
-	if ((sc_oc(sc)->flags & CF_SHUTW))
-		goto close;
-
 	/* if the connection is not established, inform the stream that we want
 	 * to be notified whenever the connection completes.
 	 */
@@ -345,7 +332,7 @@ static void sink_forward_io_handler(struct appctx *appctx)
 		applet_need_more_data(appctx);
 		se_need_remote_conn(appctx->sedesc);
 		applet_have_more_data(appctx);
-		return;
+		goto out;
 	}
 
 	HA_SPIN_LOCK(SFT_LOCK, &sft->lock);
@@ -353,7 +340,6 @@ static void sink_forward_io_handler(struct appctx *appctx)
 		HA_SPIN_UNLOCK(SFT_LOCK, &sft->lock);
 		goto close;
 	}
-	ofs = sft->ofs;
 
 	HA_RWLOCK_WRLOCK(LOGSRV_LOCK, &ring->lock);
 	LIST_DEL_INIT(&appctx->wait_entry);
@@ -369,64 +355,62 @@ static void sink_forward_io_handler(struct appctx *appctx)
 	 * existing messages before grabbing a reference to a location. This
 	 * value cannot be produced after initialization.
 	 */
-	if (unlikely(ofs == ~0)) {
-		ofs = 0;
-
-		HA_ATOMIC_INC(b_peek(buf, ofs));
-		ofs += ring->ofs;
+	if (unlikely(sft->ofs == ~0)) {
+		sft->ofs = b_peek_ofs(buf, 0);
+		HA_ATOMIC_INC(b_orig(buf) + sft->ofs);
 	}
 
+	/* we were already there, adjust the offset to be relative to
+	 * the buffer's head and remove us from the counter.
+	 */
+	ofs = sft->ofs - b_head_ofs(buf);
+	if (sft->ofs < b_head_ofs(buf))
+		ofs += b_size(buf);
+	BUG_ON(ofs >= buf->size);
+	HA_ATOMIC_DEC(b_peek(buf, ofs));
+
 	/* in this loop, ofs always points to the counter byte that precedes
 	 * the message so that we can take our reference there if we have to
 	 * stop before the end (ret=0).
 	 */
-	if (sc_opposite(sc)->state == SC_ST_EST) {
-		/* we were already there, adjust the offset to be relative to
-		 * the buffer's head and remove us from the counter.
-		 */
-		ofs -= ring->ofs;
-		BUG_ON(ofs >= buf->size);
-		HA_ATOMIC_DEC(b_peek(buf, ofs));
+	ret = 1;
+	while (ofs + 1 < b_data(buf)) {
+		cnt = 1;
+		len = b_peek_varint(buf, ofs + cnt, &msg_len);
+		if (!len)
+			break;
+		cnt += len;
+		BUG_ON(msg_len + ofs + cnt + 1 > b_data(buf));
 
-		ret = 1;
-		while (ofs + 1 < b_data(buf)) {
-			cnt = 1;
-			len = b_peek_varint(buf, ofs + cnt, &msg_len);
-			if (!len)
-				break;
-			cnt += len;
-			BUG_ON(msg_len + ofs + cnt + 1 > b_data(buf));
-
-			if (unlikely(msg_len + 1 > b_size(&trash))) {
-				/* too large a message to ever fit, let's skip it */
-				ofs += cnt + msg_len;
-				continue;
-			}
-
-			chunk_reset(&trash);
-			len = b_getblk(buf, trash.area, msg_len, ofs + cnt);
-			trash.data += len;
-			trash.area[trash.data++] = '\n';
-
-			if (applet_putchk(appctx, &trash) == -1) {
-				ret = 0;
-				break;
-			}
+		if (unlikely(msg_len + 1 > b_size(&trash))) {
+			/* too large a message to ever fit, let's skip it */
 			ofs += cnt + msg_len;
+			continue;
 		}
 
-		HA_ATOMIC_INC(b_peek(buf, ofs));
-		ofs += ring->ofs;
-		sft->ofs = ofs;
-		last_ofs = ring->ofs;
+		chunk_reset(&trash);
+		len = b_getblk(buf, trash.area, msg_len, ofs + cnt);
+		trash.data += len;
+		trash.area[trash.data++] = '\n';
+
+		if (applet_putchk(appctx, &trash) == -1) {
+			ret = 0;
+			break;
+		}
+		ofs += cnt + msg_len;
 	}
+
+	HA_ATOMIC_INC(b_peek(buf, ofs));
+	last_ofs = b_tail_ofs(buf);
+	sft->ofs = b_peek_ofs(buf, ofs);
+
 	HA_RWLOCK_RDUNLOCK(LOGSRV_LOCK, &ring->lock);
 
 	if (ret) {
 		/* let's be woken up once new data arrive */
 		HA_RWLOCK_WRLOCK(LOGSRV_LOCK, &ring->lock);
 		LIST_APPEND(&ring->waiters, &appctx->wait_entry);
-		ofs = ring->ofs;
+		ofs = b_tail_ofs(buf);
 		HA_RWLOCK_WRUNLOCK(LOGSRV_LOCK, &ring->lock);
 		if (ofs != last_ofs) {
 			/* more data was added into the ring between the
@@ -439,14 +423,13 @@ static void sink_forward_io_handler(struct appctx *appctx)
 	}
 	HA_SPIN_UNLOCK(SFT_LOCK, &sft->lock);
 
+out:
 	/* always drain data from server */
 	co_skip(sc_oc(sc), sc_oc(sc)->output);
 	return;
 
 close:
-	sc_shutw(sc);
-	sc_shutr(sc);
-	sc_ic(sc)->flags |= CF_READ_NULL;
+	se_fl_set(appctx->sedesc, SE_FL_EOS|SE_FL_EOI);
 }
 
 /*
@@ -467,26 +450,13 @@ static void sink_forward_oc_io_handler(struct appctx *appctx)
 	int ret = 0;
 	char *p;
 
+	if (unlikely(se_fl_test(appctx->sedesc, (SE_FL_EOS|SE_FL_ERROR|SE_FL_SHR|SE_FL_SHW))))
+		goto out;
+
 	/* if stopping was requested, close immediately */
 	if (unlikely(stopping))
 		goto close;
 
-	/* for rex because it seems reset to timeout
-	 * and we don't want expire on this case
-	 * with a syslog server
-	 */
-	sc_oc(sc)->rex = TICK_ETERNITY;
-	/* rto should not change but it seems the case */
-	sc_oc(sc)->rto = TICK_ETERNITY;
-
-	/* an error was detected */
-	if (unlikely(sc_ic(sc)->flags & (CF_WRITE_ERROR|CF_SHUTW)))
-		goto close;
-
-	/* con closed by server side */
-	if ((sc_oc(sc)->flags & CF_SHUTW))
-		goto close;
-
 	/* if the connection is not established, inform the stream that we want
 	 * to be notified whenever the connection completes.
 	 */
@@ -494,7 +464,7 @@ static void sink_forward_oc_io_handler(struct appctx *appctx)
 		applet_need_more_data(appctx);
 		se_need_remote_conn(appctx->sedesc);
 		applet_have_more_data(appctx);
-		return;
+		goto out;
 	}
 
 	HA_SPIN_LOCK(SFT_LOCK, &sft->lock);
@@ -502,7 +472,6 @@ static void sink_forward_oc_io_handler(struct appctx *appctx)
 		HA_SPIN_UNLOCK(SFT_LOCK, &sft->lock);
 		goto close;
 	}
-	ofs = sft->ofs;
 
 	HA_RWLOCK_WRLOCK(LOGSRV_LOCK, &ring->lock);
 	LIST_DEL_INIT(&appctx->wait_entry);
@@ -518,60 +487,58 @@ static void sink_forward_oc_io_handler(struct appctx *appctx)
 	 * existing messages before grabbing a reference to a location. This
 	 * value cannot be produced after initialization.
 	 */
-	if (unlikely(ofs == ~0)) {
-		ofs = 0;
-
-		HA_ATOMIC_INC(b_peek(buf, ofs));
-		ofs += ring->ofs;
+	if (unlikely(sft->ofs == ~0)) {
+		sft->ofs = b_peek_ofs(buf, 0);
+		HA_ATOMIC_INC(b_orig(buf) + sft->ofs);
 	}
 
+	/* we were already there, adjust the offset to be relative to
+	 * the buffer's head and remove us from the counter.
+	 */
+	ofs = sft->ofs - b_head_ofs(buf);
+	if (sft->ofs < b_head_ofs(buf))
+		ofs += b_size(buf);
+	BUG_ON(ofs >= buf->size);
+	HA_ATOMIC_DEC(b_peek(buf, ofs));
+
 	/* in this loop, ofs always points to the counter byte that precedes
 	 * the message so that we can take our reference there if we have to
 	 * stop before the end (ret=0).
 	 */
-	if (sc_opposite(sc)->state == SC_ST_EST) {
-		/* we were already there, adjust the offset to be relative to
-		 * the buffer's head and remove us from the counter.
-		 */
-		ofs -= ring->ofs;
-		BUG_ON(ofs >= buf->size);
-		HA_ATOMIC_DEC(b_peek(buf, ofs));
+	ret = 1;
+	while (ofs + 1 < b_data(buf)) {
+		cnt = 1;
+		len = b_peek_varint(buf, ofs + cnt, &msg_len);
+		if (!len)
+			break;
+		cnt += len;
+		BUG_ON(msg_len + ofs + cnt + 1 > b_data(buf));
 
-		ret = 1;
-		while (ofs + 1 < b_data(buf)) {
-			cnt = 1;
-			len = b_peek_varint(buf, ofs + cnt, &msg_len);
-			if (!len)
-				break;
-			cnt += len;
-			BUG_ON(msg_len + ofs + cnt + 1 > b_data(buf));
-
-			chunk_reset(&trash);
-			p = ulltoa(msg_len, trash.area, b_size(&trash));
-			if (p) {
-				trash.data = (p - trash.area) + 1;
-				*p = ' ';
-			}
-
-			if (!p || (trash.data + msg_len > b_size(&trash))) {
-				/* too large a message to ever fit, let's skip it */
-				ofs += cnt + msg_len;
-				continue;
-			}
-
-			trash.data += b_getblk(buf, p + 1, msg_len, ofs + cnt);
-
-			if (applet_putchk(appctx, &trash) == -1) {
-				ret = 0;
-				break;
-			}
-			ofs += cnt + msg_len;
+		chunk_reset(&trash);
+		p = ulltoa(msg_len, trash.area, b_size(&trash));
+		if (p) {
+			trash.data = (p - trash.area) + 1;
+			*p = ' ';
 		}
 
-		HA_ATOMIC_INC(b_peek(buf, ofs));
-		ofs += ring->ofs;
-		sft->ofs = ofs;
+		if (!p || (trash.data + msg_len > b_size(&trash))) {
+			/* too large a message to ever fit, let's skip it */
+			ofs += cnt + msg_len;
+			continue;
+		}
+
+		trash.data += b_getblk(buf, p + 1, msg_len, ofs + cnt);
+
+		if (applet_putchk(appctx, &trash) == -1) {
+			ret = 0;
+			break;
+		}
+		ofs += cnt + msg_len;
 	}
+
+	HA_ATOMIC_INC(b_peek(buf, ofs));
+	sft->ofs = b_peek_ofs(buf, ofs);
+
 	HA_RWLOCK_RDUNLOCK(LOGSRV_LOCK, &ring->lock);
 
 	if (ret) {
@@ -583,14 +550,14 @@ static void sink_forward_oc_io_handler(struct appctx *appctx)
 	}
 	HA_SPIN_UNLOCK(SFT_LOCK, &sft->lock);
 
+  out:
 	/* always drain data from server */
 	co_skip(sc_oc(sc), sc_oc(sc)->output);
 	return;
 
 close:
-	sc_shutw(sc);
-	sc_shutr(sc);
-	sc_ic(sc)->flags |= CF_READ_NULL;
+	se_fl_set(appctx->sedesc, SE_FL_EOS|SE_FL_EOI);
+	goto out;
 }
 
 void __sink_forward_session_deinit(struct sink_forward_target *sft)
@@ -624,7 +591,7 @@ static int sink_forward_session_init(struct appctx *appctx)
 
 	s = appctx_strm(appctx);
 	s->scb->dst = addr;
-	s->scb->flags |= SC_FL_NOLINGER;
+	s->scb->flags |= (SC_FL_RCV_ONCE|SC_FL_NOLINGER);
 
 	s->target = &sft->srv->obj_type;
 	s->flags = SF_ASSIGNED;
@@ -632,12 +599,7 @@ static int sink_forward_session_init(struct appctx *appctx)
 	s->do_log = NULL;
 	s->uniq_id = 0;
 
-	s->res.flags |= CF_READ_DONTWAIT;
-	/* for rto and rex to eternity to not expire on idle recv:
-	 * We are using a syslog server.
-	 */
-	s->res.rto = TICK_ETERNITY;
-	s->res.rex = TICK_ETERNITY;
+	applet_expect_no_data(appctx);
 	sft->appctx = appctx;
 
 	return 0;
@@ -1202,8 +1164,7 @@ error:
 		free(sft);
 
 	if (sink) {
-		if (sink->ctx.ring)
-			ring_free(sink->ctx.ring);
+		ring_free(sink->ctx.ring);
 
 		LIST_DELETE(&sink->sink_list);
 		free(sink->name);
