@@ -21,12 +21,12 @@
 #include <import/ist.h>
 #include <haproxy/api.h>
 #include <haproxy/buf.h>
+#include <haproxy/cfgparse.h>
 #include <haproxy/cli.h>
 #include <haproxy/errors.h>
 #include <haproxy/istbuf.h>
 #include <haproxy/list.h>
 #include <haproxy/log.h>
-#include <haproxy/quic_conn-t.h>
 #include <haproxy/sink.h>
 #include <haproxy/trace.h>
 
@@ -98,6 +98,7 @@ int __trace_enabled(enum trace_level level, uint64_t mask, struct trace_source *
 	const struct connection *conn = NULL;
 	const struct check *check = NULL;
 	const struct quic_conn *qc = NULL;
+	const struct appctx *appctx = NULL;
 	const void *lockon_ptr = NULL;
 
 	if (likely(src->state == TRACE_STATE_STOPPED))
@@ -123,16 +124,17 @@ int __trace_enabled(enum trace_level level, uint64_t mask, struct trace_source *
 	if (src->arg_def & TRC_ARGS_QCON)
 		qc = trace_pick_arg(src->arg_def & TRC_ARGS_QCON, a1, a2, a3, a4);
 
-#ifdef USE_QUIC
-	if (qc && !conn)
-		conn = qc->conn;
-#endif
+	if (src->arg_def & TRC_ARGS_APPCTX)
+		appctx = trace_pick_arg(src->arg_def & TRC_ARGS_APPCTX, a1, a2, a3, a4);
+
 	if (!sess && strm)
 		sess = strm->sess;
-	else if (!sess && conn && LIST_INLIST(&conn->sess_el))
+	else if (!sess && conn && LIST_INLIST(&conn->session_list))
 		sess = conn->owner;
 	else if (!sess && check)
 		sess = check->sess;
+	else if (!sess && appctx)
+		sess = appctx->sess;
 
 	if (sess) {
 		fe = sess->fe;
@@ -177,20 +179,21 @@ int __trace_enabled(enum trace_level level, uint64_t mask, struct trace_source *
 	/* we may want to lock on a particular object */
 	if (src->lockon != TRACE_LOCKON_NOTHING) {
 		switch (src->lockon) {
-		case TRACE_LOCKON_BACKEND:    lockon_ptr = be;   break;
-		case TRACE_LOCKON_CONNECTION: lockon_ptr = conn; break;
-		case TRACE_LOCKON_FRONTEND:   lockon_ptr = fe;   break;
-		case TRACE_LOCKON_LISTENER:   lockon_ptr = li;   break;
-		case TRACE_LOCKON_SERVER:     lockon_ptr = srv;  break;
-		case TRACE_LOCKON_SESSION:    lockon_ptr = sess; break;
-		case TRACE_LOCKON_STREAM:     lockon_ptr = strm; break;
-		case TRACE_LOCKON_CHECK:      lockon_ptr = check; break;
-		case TRACE_LOCKON_THREAD:     lockon_ptr = ti;   break;
-		case TRACE_LOCKON_QCON:       lockon_ptr = qc;   break;
-		case TRACE_LOCKON_ARG1:       lockon_ptr = a1;   break;
-		case TRACE_LOCKON_ARG2:       lockon_ptr = a2;   break;
-		case TRACE_LOCKON_ARG3:       lockon_ptr = a3;   break;
-		case TRACE_LOCKON_ARG4:       lockon_ptr = a4;   break;
+		case TRACE_LOCKON_BACKEND:    lockon_ptr = be;     break;
+		case TRACE_LOCKON_CONNECTION: lockon_ptr = conn;   break;
+		case TRACE_LOCKON_FRONTEND:   lockon_ptr = fe;     break;
+		case TRACE_LOCKON_LISTENER:   lockon_ptr = li;     break;
+		case TRACE_LOCKON_SERVER:     lockon_ptr = srv;    break;
+		case TRACE_LOCKON_SESSION:    lockon_ptr = sess;   break;
+		case TRACE_LOCKON_STREAM:     lockon_ptr = strm;   break;
+		case TRACE_LOCKON_CHECK:      lockon_ptr = check;  break;
+		case TRACE_LOCKON_THREAD:     lockon_ptr = ti;     break;
+		case TRACE_LOCKON_QCON:       lockon_ptr = qc;     break;
+		case TRACE_LOCKON_APPCTX:     lockon_ptr = appctx; break;
+		case TRACE_LOCKON_ARG1:       lockon_ptr = a1;     break;
+		case TRACE_LOCKON_ARG2:       lockon_ptr = a2;     break;
+		case TRACE_LOCKON_ARG3:       lockon_ptr = a3;     break;
+		case TRACE_LOCKON_ARG4:       lockon_ptr = a4;     break;
 		default: break; // silence stupid gcc -Wswitch
 		}
 
@@ -288,7 +291,7 @@ void __trace(enum trace_level level, uint64_t mask, struct trace_source *src,
 	}
 
 	if (src->sink)
-		sink_write(src->sink, 0, line, words, 0, 0, NULL);
+		sink_write(src->sink, line, words, 0, 0, NULL);
 
  end:
 	/* check if we need to stop the trace now */
@@ -344,14 +347,21 @@ const struct trace_event *trace_find_event(const struct trace_event *ev, const c
 	return NULL;
 }
 
-/* parse the command, returns 1 if a message is returned, otherwise zero */
-static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, void *private)
+/* Parse a "trace" statement. Returns a severity as a LOG_* level and a status
+ * message that may be delivered to the user, in <msg>. The message will be
+ * nulled first and msg must be a valid pointer. A null status message output
+ * indicates no error. Be careful not to use the return value as a boolean, as
+ * LOG_* values are not ordered as one could imagine (LOG_EMERG is zero). The
+ * function may/will use the trash buffer as the storage for the response
+ * message so that the caller never needs to release anything.
+ */
+static int trace_parse_statement(char **args, const char **msg)
 {
 	struct trace_source *src;
 	uint64_t *ev_ptr = NULL;
 
-	if (!cli_has_level(appctx, ACCESS_LVL_OPER))
-		return 1;
+	/* no error by default */
+	*msg = NULL;
 
 	if (!*args[1]) {
 		/* no arg => report the list of supported sources as a warning */
@@ -364,33 +374,36 @@ static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, vo
 			chunk_appendf(&trash, " [%c] %-10s : %s\n", trace_state_char(src->state), src->name.ptr, src->desc);
 
 		trash.area[trash.data] = 0;
-		return cli_msg(appctx, LOG_WARNING, trash.area);
+		*msg = trash.area;
+		return LOG_WARNING;
 	}
 
 	if (strcmp(args[1], "0") == 0) {
 		/* emergency stop of all traces */
 		list_for_each_entry(src, &trace_sources, source_link)
 			HA_ATOMIC_STORE(&src->state, TRACE_STATE_STOPPED);
-		return cli_msg(appctx, LOG_NOTICE, "All traces now stopped");
+		*msg = "All traces now stopped";
+		return LOG_NOTICE;
 	}
 
 	src = trace_find_source(args[1]);
-	if (!src)
-		return cli_err(appctx, "No such trace source");
+	if (!src) {
+		*msg = "No such trace source";
+		return LOG_ERR;
+	}
 
 	if (!*args[2]) {
-		return cli_msg(appctx, LOG_WARNING,
-			       "Supported commands:\n"
-			       "  event     : list/enable/disable source-specific event reporting\n"
-			       //"  filter    : list/enable/disable generic filters\n"
-			       "  level     : list/set trace reporting level\n"
-			       "  lock      : automatic lock on thread/connection/stream/...\n"
-			       "  pause     : pause and automatically restart after a specific event\n"
-			       "  sink      : list/set event sinks\n"
-			       "  start     : start immediately or after a specific event\n"
-			       "  stop      : stop immediately or after a specific event\n"
-			       "  verbosity : list/set trace output verbosity\n"
-			       );
+		*msg =  "Supported commands:\n"
+			"  event     : list/enable/disable source-specific event reporting\n"
+			//"  filter    : list/enable/disable generic filters\n"
+			"  level     : list/set trace reporting level\n"
+			"  lock      : automatic lock on thread/connection/stream/...\n"
+			"  pause     : pause and automatically restart after a specific event\n"
+			"  sink      : list/set event sinks\n"
+			"  start     : start immediately or after a specific event\n"
+			"  stop      : stop immediately or after a specific event\n"
+			"  verbosity : list/set trace output verbosity\n";
+		return LOG_WARNING;
 	}
 	else if ((strcmp(args[2], "event") == 0 && (ev_ptr = &src->report_events)) ||
 	         (strcmp(args[2], "pause") == 0 && (ev_ptr = &src->pause_events)) ||
@@ -424,16 +437,10 @@ static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, vo
 					      src->known_events[i].name, src->known_events[i].desc);
 			}
 			trash.area[trash.data] = 0;
-			return cli_msg(appctx, LOG_WARNING, trash.area);
+			*msg = trash.area;
+			return LOG_WARNING;
 		}
 
-		/* state transitions:
-		 *   - "start now" => TRACE_STATE_RUNNING
-		 *   - "stop now"  => TRACE_STATE_STOPPED
-		 *   - "pause now" => TRACE_STATE_WAITING
-		 *   - "start <evt>" && STATE_STOPPED => TRACE_STATE_WAITING
-		 */
-
 		if (strcmp(name, "now") == 0 && ev_ptr != &src->report_events) {
 			HA_ATOMIC_STORE(ev_ptr, 0);
 			if (ev_ptr == &src->pause_events) {
@@ -447,32 +454,25 @@ static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, vo
 				HA_ATOMIC_STORE(&src->lockon_ptr, NULL);
 				HA_ATOMIC_STORE(&src->state, TRACE_STATE_STOPPED);
 			}
+			*msg = NULL;
 			return 0;
 		}
 
 		if (strcmp(name, "none") == 0)
 			HA_ATOMIC_STORE(ev_ptr, 0);
-		else if (strcmp(name, "any") == 0) {
-			enum trace_state old = TRACE_STATE_STOPPED;
-
+		else if (strcmp(name, "any") == 0)
 			HA_ATOMIC_STORE(ev_ptr, ~0);
-			if (ev_ptr == &src->start_events)
-				HA_ATOMIC_CAS(&src->state, &old, TRACE_STATE_WAITING);
-		}
 		else {
-			enum trace_state old = TRACE_STATE_STOPPED;
-
 			ev = trace_find_event(src->known_events, name);
-			if (!ev)
-				return cli_err(appctx, "No such trace event");
+			if (!ev) {
+				*msg = "No such trace event";
+				return LOG_ERR;
+			}
 
 			if (!neg)
 				HA_ATOMIC_OR(ev_ptr, ev->mask);
 			else
 				HA_ATOMIC_AND(ev_ptr, ~ev->mask);
-
-			if (ev_ptr == &src->start_events && HA_ATOMIC_LOAD(ev_ptr) != 0)
-				HA_ATOMIC_CAS(&src->state, &old, TRACE_STATE_WAITING);
 		}
 	}
 	else if (strcmp(args[2], "sink") == 0) {
@@ -488,15 +488,18 @@ static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, vo
 					      sink->name, sink->desc);
 			}
 			trash.area[trash.data] = 0;
-			return cli_msg(appctx, LOG_WARNING, trash.area);
+			*msg = trash.area;
+			return LOG_WARNING;
 		}
 
 		if (strcmp(name, "none") == 0)
 			sink = NULL;
 		else {
 			sink = sink_find(name);
-			if (!sink)
-				return cli_err(appctx, "No such sink");
+			if (!sink) {
+				*msg = "No such sink";
+				return LOG_ERR;
+			}
 		}
 
 		HA_ATOMIC_STORE(&src->sink, sink);
@@ -519,7 +522,8 @@ static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, vo
 			chunk_appendf(&trash, "  %c developer  : also report information useful only to the developer\n",
 				      src->level == TRACE_LEVEL_DEVELOPER ? '*' : ' ');
 			trash.area[trash.data] = 0;
-			return cli_msg(appctx, LOG_WARNING, trash.area);
+			*msg = trash.area;
+			return LOG_WARNING;
 		}
 
 		if (strcmp(name, "error") == 0)
@@ -534,8 +538,10 @@ static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, vo
 			HA_ATOMIC_STORE(&src->level, TRACE_LEVEL_DATA);
 		else if (strcmp(name, "developer") == 0)
 			HA_ATOMIC_STORE(&src->level, TRACE_LEVEL_DEVELOPER);
-		else
-			return cli_err(appctx, "No such trace level");
+		else {
+			*msg = "No such trace level";
+			return LOG_ERR;
+		}
 	}
 	else if (strcmp(args[2], "lock") == 0) {
 		const char *name = args[3];
@@ -550,15 +556,15 @@ static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, vo
 				chunk_appendf(&trash, "  %c check      : lock on the check that started the trace\n",
 				              src->lockon == TRACE_LOCKON_CHECK ? '*' : ' ');
 
-			if (src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_QCON))
+			if (src->arg_def & TRC_ARGS_CONN)
 				chunk_appendf(&trash, "  %c connection : lock on the connection that started the trace\n",
 				              src->lockon == TRACE_LOCKON_CONNECTION ? '*' : ' ');
 
-			if (src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_QCON|TRC_ARGS_SESS|TRC_ARGS_STRM))
+			if (src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_SESS|TRC_ARGS_STRM))
 				chunk_appendf(&trash, "  %c frontend   : lock on the frontend that started the trace\n",
 				              src->lockon == TRACE_LOCKON_FRONTEND ? '*' : ' ');
 
-			if (src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_QCON|TRC_ARGS_SESS|TRC_ARGS_STRM))
+			if (src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_SESS|TRC_ARGS_STRM))
 				chunk_appendf(&trash, "  %c listener   : lock on the listener that started the trace\n",
 				              src->lockon == TRACE_LOCKON_LISTENER ? '*' : ' ');
 
@@ -568,12 +574,8 @@ static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, vo
 			if (src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_STRM))
 				chunk_appendf(&trash, "  %c server     : lock on the server that started the trace\n",
 				              src->lockon == TRACE_LOCKON_SERVER ? '*' : ' ');
-#ifdef USE_QUIC
-			if (src->arg_def & TRC_ARGS_QCON)
-				chunk_appendf(&trash, "  %c qconn      : lock on the QUIC connection that started the trace\n",
-				              src->lockon == TRACE_LOCKON_QCON ? '*' : ' ');
-#endif
-			if (src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_QCON|TRC_ARGS_SESS|TRC_ARGS_STRM))
+
+			if (src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_SESS|TRC_ARGS_STRM))
 				chunk_appendf(&trash, "  %c session    : lock on the session that started the trace\n",
 				              src->lockon == TRACE_LOCKON_SESSION ? '*' : ' ');
 
@@ -581,6 +583,10 @@ static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, vo
 				chunk_appendf(&trash, "  %c stream     : lock on the stream that started the trace\n",
 				              src->lockon == TRACE_LOCKON_STREAM ? '*' : ' ');
 
+			if (src->arg_def & TRC_ARGS_APPCTX)
+				chunk_appendf(&trash, "  %c applet     : lock on the applet that started the trace\n",
+				              src->lockon == TRACE_LOCKON_APPCTX ? '*' : ' ');
+
 			chunk_appendf(&trash, "  %c thread     : lock on the thread that started the trace\n",
 				      src->lockon == TRACE_LOCKON_THREAD ? '*' : ' ');
 
@@ -605,7 +611,8 @@ static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, vo
 				              src->lockon_args[3].name, src->lockon_args[3].desc);
 
 			trash.area[trash.data] = 0;
-			return cli_msg(appctx, LOG_WARNING, trash.area);
+			*msg = trash.area;
+			return LOG_WARNING;
 		}
 		else if ((src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_STRM)) && strcmp(name, "backend") == 0) {
 			HA_ATOMIC_STORE(&src->lockon, TRACE_LOCKON_BACKEND);
@@ -615,15 +622,15 @@ static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, vo
 			HA_ATOMIC_STORE(&src->lockon, TRACE_LOCKON_CHECK);
 			HA_ATOMIC_STORE(&src->lockon_ptr, NULL);
 		}
-		else if ((src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_QCON)) && strcmp(name, "connection") == 0) {
+		else if ((src->arg_def & TRC_ARGS_CONN) && strcmp(name, "connection") == 0) {
 			HA_ATOMIC_STORE(&src->lockon, TRACE_LOCKON_CONNECTION);
 			HA_ATOMIC_STORE(&src->lockon_ptr, NULL);
 		}
-		else if ((src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_QCON|TRC_ARGS_SESS|TRC_ARGS_STRM)) && strcmp(name, "frontend") == 0) {
+		else if ((src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_SESS|TRC_ARGS_STRM)) && strcmp(name, "frontend") == 0) {
 			HA_ATOMIC_STORE(&src->lockon, TRACE_LOCKON_FRONTEND);
 			HA_ATOMIC_STORE(&src->lockon_ptr, NULL);
 		}
-		else if ((src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_QCON|TRC_ARGS_SESS|TRC_ARGS_STRM)) && strcmp(name, "listener") == 0) {
+		else if ((src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_SESS|TRC_ARGS_STRM)) && strcmp(name, "listener") == 0) {
 			HA_ATOMIC_STORE(&src->lockon, TRACE_LOCKON_LISTENER);
 			HA_ATOMIC_STORE(&src->lockon_ptr, NULL);
 		}
@@ -635,18 +642,18 @@ static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, vo
 			HA_ATOMIC_STORE(&src->lockon, TRACE_LOCKON_SERVER);
 			HA_ATOMIC_STORE(&src->lockon_ptr, NULL);
 		}
-		else if ((src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_QCON|TRC_ARGS_SESS|TRC_ARGS_STRM)) && strcmp(name, "session") == 0) {
+		else if ((src->arg_def & (TRC_ARGS_CONN|TRC_ARGS_SESS|TRC_ARGS_STRM)) && strcmp(name, "session") == 0) {
 			HA_ATOMIC_STORE(&src->lockon, TRACE_LOCKON_SESSION);
 			HA_ATOMIC_STORE(&src->lockon_ptr, NULL);
 		}
-		else if ((src->arg_def & TRC_ARGS_QCON) && strcmp(name, "qconn") == 0) {
-			HA_ATOMIC_STORE(&src->lockon, TRACE_LOCKON_QCON);
-			HA_ATOMIC_STORE(&src->lockon_ptr, NULL);
-		}
 		else if ((src->arg_def & TRC_ARGS_STRM) && strcmp(name, "stream") == 0) {
 			HA_ATOMIC_STORE(&src->lockon, TRACE_LOCKON_STREAM);
 			HA_ATOMIC_STORE(&src->lockon_ptr, NULL);
 		}
+		else if ((src->arg_def & TRC_ARGS_APPCTX) && strcmp(name, "appctx") == 0) {
+			HA_ATOMIC_STORE(&src->lockon, TRACE_LOCKON_APPCTX);
+			HA_ATOMIC_STORE(&src->lockon_ptr, NULL);
+		}
 		else if (strcmp(name, "thread") == 0) {
 			HA_ATOMIC_STORE(&src->lockon, TRACE_LOCKON_THREAD);
 			HA_ATOMIC_STORE(&src->lockon_ptr, NULL);
@@ -667,8 +674,10 @@ static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, vo
 			HA_ATOMIC_STORE(&src->lockon, TRACE_LOCKON_ARG4);
 			HA_ATOMIC_STORE(&src->lockon_ptr, NULL);
 		}
-		else
-			return cli_err(appctx, "Unsupported lock-on criterion");
+		else {
+			*msg = "Unsupported lock-on criterion";
+			return LOG_ERR;
+		}
 	}
 	else if (strcmp(args[2], "verbosity") == 0) {
 		const char *name = args[3];
@@ -688,7 +697,8 @@ static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, vo
 						      nd->name, nd->desc);
 			}
 			trash.area[trash.data] = 0;
-			return cli_msg(appctx, LOG_WARNING, trash.area);
+			*msg = trash.area;
+			return LOG_WARNING;
 		}
 
 		if (strcmp(name, "quiet") == 0)
@@ -696,22 +706,67 @@ static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, vo
 		else if (!src->decoding || !src->decoding[0].name) {
 			if (strcmp(name, "default") == 0)
 				HA_ATOMIC_STORE(&src->verbosity, 1);
-			else
-				return cli_err(appctx, "No such verbosity level");
+			else {
+				*msg = "No such verbosity level";
+				return LOG_ERR;
+			}
 		} else {
 			for (nd = src->decoding; nd->name && nd->desc; nd++)
 				if (strcmp(name, nd->name) == 0)
 					break;
 
-			if (!nd->name || !nd->desc)
-				return cli_err(appctx, "No such verbosity level");
+			if (!nd->name || !nd->desc) {
+				*msg = "No such verbosiry level";
+				return LOG_ERR;
+			}
 
 			HA_ATOMIC_STORE(&src->verbosity, (nd - src->decoding) + 1);
 		}
 	}
-	else
-		return cli_err(appctx, "Unknown trace keyword");
+	else {
+		*msg = "Unknown trace keyword";
+		return LOG_ERR;
+	}
+	return 0;
 
+}
+
+/* parse a "trace" statement in the "global" section, returns 1 if a message is returned, otherwise zero */
+static int cfg_parse_trace(char **args, int section_type, struct proxy *curpx,
+			   const struct proxy *defpx, const char *file, int line,
+			   char **err)
+{
+	const char *msg;
+	int severity;
+
+	severity = trace_parse_statement(args, &msg);
+	if (msg) {
+		if (severity >= LOG_NOTICE)
+			ha_notice("parsing [%s:%d] : '%s': %s\n", file, line, args[0], msg);
+		else if (severity >= LOG_WARNING)
+			ha_warning("parsing [%s:%d] : '%s': %s\n", file, line, args[0], msg);
+		else {
+			ha_alert("parsing [%s:%d] : '%s': %s\n", file, line, args[0], msg);
+			return -1;
+		}
+	}
+	return 0;
+}
+
+/* parse the command, returns 1 if a message is returned, otherwise zero */
+static int cli_parse_trace(char **args, char *payload, struct appctx *appctx, void *private)
+{
+	const char *msg;
+	int severity;
+
+	if (!cli_has_level(appctx, ACCESS_LVL_OPER))
+		return 1;
+
+	severity = trace_parse_statement(args, &msg);
+	if (msg)
+		return cli_msg(appctx, severity, msg);
+
+	/* total success */
 	return 0;
 }
 
@@ -777,6 +832,13 @@ static struct cli_kw_list cli_kws = {{ },{
 
 INITCALL1(STG_REGISTER, cli_register_kw, &cli_kws);
 
+static struct cfg_kw_list cfg_kws = {ILH, {
+	{ CFG_GLOBAL, "trace", cfg_parse_trace, KWF_EXPERIMENTAL },
+	{ /* END */ },
+}};
+
+INITCALL1(STG_REGISTER, cfg_register_keywords, &cfg_kws);
+
 /*
  * Local variables:
  *  c-indent-level: 8
